import copy
import re

import pandas as pd
import pytest
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime

from featuretools import EntitySet, Feature, GroupByTransformFeature
from featuretools.entityset.entityset import LTI_COLUMN_NAME
from featuretools.feature_base import (
    AggregationFeature,
    DirectFeature,
    IdentityFeature,
    TransformFeature,
)
from featuretools.feature_base.utils import is_valid_input
from featuretools.primitives import (
    Absolute,
    AddNumeric,
    Count,
    CumCount,
    CumMean,
    CumMin,
    CumSum,
    Day,
    Diff,
    Equal,
    Hour,
    IsIn,
    IsNull,
    Last,
    Mean,
    Mode,
    Month,
    Negate,
    NMostCommon,
    Not,
    NotEqual,
    NumCharacters,
    NumTrue,
    NumUnique,
    RollingCount,
    RollingMax,
    RollingMean,
    RollingMin,
    RollingOutlierCount,
    RollingSTD,
    Sum,
    TimeSincePrevious,
    TransformPrimitive,
    Trend,
    Year,
)
from featuretools.synthesis import DeepFeatureSynthesis
from featuretools.tests.testing_utils import (
    feature_with_name,
    make_ecommerce_entityset,
    number_of_features_with_name_like,
)
from featuretools.utils.gen_utils import Library


def test_makes_agg_features_from_str(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=["sum"],
        trans_primitives=[],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "SUM(log.value)")


def test_makes_agg_features_from_mixed_str(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Count, "sum"],
        trans_primitives=[],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "SUM(log.value)")
    assert feature_with_name(features, "COUNT(log)")


def test_makes_agg_features(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Sum],
        trans_primitives=[],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "SUM(log.value)")


def test_only_makes_supplied_agg_feat(es):
    kwargs = dict(
        target_dataframe_name="customers",
        entityset=es,
        max_depth=3,
    )
    dfs_obj = DeepFeatureSynthesis(agg_primitives=[Sum], **kwargs)

    features = dfs_obj.build_features()

    def find_other_agg_features(features):
        return [
            f
            for f in features
            if (isinstance(f, AggregationFeature) and not isinstance(f.primitive, Sum))
            or len(
                [
                    g
                    for g in f.base_features
                    if isinstance(g, AggregationFeature)
                    and not isinstance(g.primitive, Sum)
                ],
            )
            > 0
        ]

    other_agg_features = find_other_agg_features(features)
    assert len(other_agg_features) == 0


def test_errors_unsupported_primitives(es):
    bad_trans_prim = CumSum()
    bad_agg_prim = NumUnique()
    bad_trans_prim.compatibility, bad_agg_prim.compatibility = [], []
    library = es.dataframe_type
    error_text = "Selected primitives are incompatible with {} EntitySets: cum_sum, num_unique".format(
        library.value,
    )
    with pytest.raises(ValueError, match=error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="sessions",
            entityset=es,
            agg_primitives=[bad_agg_prim],
            trans_primitives=[bad_trans_prim],
        )


def test_errors_unsupported_primitives_spark(spark_es):
    bad_trans_prim = CumSum()
    bad_agg_prim = NumUnique()
    bad_trans_prim.spark_compatible, bad_agg_prim.spark_compatible = False, False
    error_text = "Selected primitives are incompatible with Spark EntitySets: cum_sum"
    with pytest.raises(ValueError, match=error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="sessions",
            entityset=spark_es,
            agg_primitives=[bad_agg_prim],
            trans_primitives=[bad_trans_prim],
        )


def test_error_for_missing_target_dataframe(es):
    error_text = (
        "Provided target dataframe missing_dataframe does not exist in ecommerce"
    )
    with pytest.raises(KeyError, match=error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="missing_dataframe",
            entityset=es,
            agg_primitives=[Last],
            trans_primitives=[],
            ignore_dataframes=["log"],
        )

    es_without_id = EntitySet()
    error_text = (
        "Provided target dataframe missing_dataframe does not exist in entity set"
    )
    with pytest.raises(KeyError, match=error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="missing_dataframe",
            entityset=es_without_id,
            agg_primitives=[Last],
            trans_primitives=[],
            ignore_dataframes=["log"],
        )


def test_ignores_dataframes(es):
    error_text = "ignore_dataframes must be a list"
    with pytest.raises(TypeError, match=error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="sessions",
            entityset=es,
            agg_primitives=[Sum],
            trans_primitives=[],
            ignore_dataframes="log",
        )

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Sum],
        trans_primitives=[],
        ignore_dataframes=["log"],
    )

    features = dfs_obj.build_features()
    for f in features:
        deps = f.get_dependencies(deep=True)
        dataframes = [d.dataframe_name for d in deps]
        assert "log" not in dataframes


def test_ignores_columns(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Sum],
        trans_primitives=[],
        ignore_columns={"log": ["value"]},
    )
    features = dfs_obj.build_features()
    for f in features:
        deps = f.get_dependencies(deep=True)
        identities = [d for d in deps if isinstance(d, IdentityFeature)]
        columns = [d.column_name for d in identities if d.dataframe_name == "log"]
        assert "value" not in columns


def test_ignore_columns_input_type(es):
    error_msg = r"ignore_columns should be dict\[str -> list\]"  # need to use string literals to avoid regex params
    wrong_input_type = {"log": "value"}
    with pytest.raises(TypeError, match=error_msg):
        DeepFeatureSynthesis(
            target_dataframe_name="log",
            entityset=es,
            ignore_columns=wrong_input_type,
        )


def test_ignore_columns_with_nonstring_values(es):
    error_msg = "list in ignore_columns must only have string values"
    wrong_input_list = {"log": ["a", "b", 3]}
    with pytest.raises(TypeError, match=error_msg):
        DeepFeatureSynthesis(
            target_dataframe_name="log",
            entityset=es,
            ignore_columns=wrong_input_list,
        )


def test_ignore_columns_with_nonstring_keys(es):
    error_msg = r"ignore_columns should be dict\[str -> list\]"  # need to use string literals to avoid regex params
    wrong_input_keys = {1: ["a", "b", "c"]}
    with pytest.raises(TypeError, match=error_msg):
        DeepFeatureSynthesis(
            target_dataframe_name="log",
            entityset=es,
            ignore_columns=wrong_input_keys,
        )


def test_makes_dfeatures(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[],
        trans_primitives=[],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "customers.age")


def test_makes_trans_feat(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=es,
        agg_primitives=[],
        trans_primitives=[Hour],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "HOUR(datetime)")


def test_handles_diff_dataframe_groupby(pd_es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[],
        groupby_trans_primitives=[Diff],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "DIFF(value) by session_id")
    assert feature_with_name(features, "DIFF(value) by product_id")


def test_handles_time_since_previous_dataframe_groupby(pd_es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[],
        groupby_trans_primitives=[TimeSincePrevious],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "TIME_SINCE_PREVIOUS(datetime) by session_id")


# M TODO
# def test_handles_cumsum_dataframe_groupby(pd_es):
#     dfs_obj = DeepFeatureSynthesis(target_dataframe_name='sessions',
#                                    entityset=pd_es,
#                                    agg_primitives=[],
#                                    trans_primitives=[CumMean])

#     features = dfs_obj.build_features()
#     assert (feature_with_name(features, u'customers.CUM_MEAN(age by région_id)'))


def test_only_makes_supplied_trans_feat(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=es,
        agg_primitives=[],
        trans_primitives=[Hour],
    )

    features = dfs_obj.build_features()
    other_trans_features = [
        f
        for f in features
        if (isinstance(f, TransformFeature) and not isinstance(f.primitive, Hour))
        or len(
            [
                g
                for g in f.base_features
                if isinstance(g, TransformFeature) and not isinstance(g.primitive, Hour)
            ],
        )
        > 0
    ]
    assert len(other_trans_features) == 0


def test_makes_dfeatures_of_agg_primitives(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=["max"],
        trans_primitives=[],
    )
    features = dfs_obj.build_features()

    assert feature_with_name(features, "customers.MAX(log.value)")


def test_makes_agg_features_of_trans_primitives(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Mean],
        trans_primitives=[NumCharacters],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "MEAN(log.NUM_CHARACTERS(comments))")


def test_makes_agg_features_with_where(es):
    # TODO: Update to work with Dask and Spark `es` fixture when issue #978 is closed
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Dask EntitySets do not support add_interesting_values")
    es.add_interesting_values()

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Count],
        where_primitives=[Count],
        trans_primitives=[],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "COUNT(log WHERE priority_level = 0)")

    # make sure they are made using direct features too
    assert feature_with_name(features, "COUNT(log WHERE products.department = food)")


def test_make_groupby_features(pd_es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[],
        trans_primitives=[],
        groupby_trans_primitives=["cum_sum"],
    )
    features = dfs_obj.build_features()
    assert feature_with_name(features, "CUM_SUM(value) by session_id")


def test_make_indirect_groupby_features(pd_es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[],
        trans_primitives=[],
        groupby_trans_primitives=["cum_sum"],
    )
    features = dfs_obj.build_features()
    assert feature_with_name(features, "CUM_SUM(products.rating) by session_id")


def test_make_groupby_features_with_id(pd_es):
    # Need to convert customer_id to categorical column in order to build desired feature
    pd_es["sessions"].ww.set_types(
        logical_types={"customer_id": "Categorical"},
        semantic_tags={"customer_id": "foreign_key"},
    )
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=pd_es,
        agg_primitives=[],
        trans_primitives=[],
        groupby_trans_primitives=["cum_count"],
    )
    features = dfs_obj.build_features()

    assert feature_with_name(features, "CUM_COUNT(customer_id) by customer_id")


def test_make_groupby_features_with_diff_id(pd_es):
    # Need to convert cohort to categorical column in order to build desired feature
    pd_es["customers"].ww.set_types(
        logical_types={"cohort": "Categorical"},
        semantic_tags={"cohort": "foreign_key"},
    )
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=pd_es,
        agg_primitives=[],
        trans_primitives=[],
        groupby_trans_primitives=["cum_count"],
    )
    features = dfs_obj.build_features()

    groupby_with_diff_id = "CUM_COUNT(cohort) by région_id"
    assert feature_with_name(features, groupby_with_diff_id)


def test_make_groupby_features_with_agg(pd_es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="cohorts",
        entityset=pd_es,
        agg_primitives=["sum"],
        trans_primitives=[],
        groupby_trans_primitives=["cum_sum"],
    )
    features = dfs_obj.build_features()
    agg_on_groupby_name = "SUM(customers.CUM_SUM(age) by région_id)"
    assert feature_with_name(features, agg_on_groupby_name)


def test_bad_groupby_feature(es):
    msg = re.escape(
        "Unknown groupby transform primitive max. "
        "Call ft.primitives.list_primitives() to get "
        "a list of available primitives",
    )
    with pytest.raises(ValueError, match=msg):
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=["sum"],
            trans_primitives=[],
            groupby_trans_primitives=["Max"],
        )


@pytest.mark.parametrize(
    "rolling_primitive",
    [
        RollingMax,
        RollingMean,
        RollingMin,
        RollingOutlierCount,
        RollingSTD,
    ],
)
@pytest.mark.parametrize(
    "window_length, gap",
    [
        (7, 3),
        ("7d", "3d"),
    ],
)
def test_make_rolling_features(window_length, gap, rolling_primitive, pd_es):
    rolling_primitive_obj = rolling_primitive(
        window_length=window_length,
        gap=gap,
        min_periods=5,
    )
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[],
        trans_primitives=[rolling_primitive_obj],
    )
    features = dfs_obj.build_features()
    rolling_transform_name = f"{rolling_primitive.name.upper()}(datetime, value_many_nans, window_length={window_length}, gap={gap}, min_periods=5)"
    assert feature_with_name(features, rolling_transform_name)


@pytest.mark.parametrize(
    "window_length, gap",
    [
        (7, 3),
        ("7d", "3d"),
    ],
)
def test_make_rolling_count_off_datetime_feature(window_length, gap, pd_es):
    rolling_count = RollingCount(window_length=window_length, min_periods=gap)
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[],
        trans_primitives=[rolling_count],
    )
    features = dfs_obj.build_features()
    rolling_transform_name = (
        f"ROLLING_COUNT(datetime, window_length={window_length}, min_periods={gap})"
    )
    assert feature_with_name(features, rolling_transform_name)


def test_abides_by_max_depth_param(es):
    for i in [0, 1, 2, 3]:
        dfs_obj = DeepFeatureSynthesis(
            target_dataframe_name="sessions",
            entityset=es,
            agg_primitives=[Sum],
            trans_primitives=[],
            max_depth=i,
        )

        features = dfs_obj.build_features()
        for f in features:
            assert f.get_depth() <= i


def test_max_depth_single_table(transform_es):
    assert len(transform_es.dataframe_dict) == 1

    def make_dfs_obj(max_depth):
        dfs_obj = DeepFeatureSynthesis(
            target_dataframe_name="first",
            entityset=transform_es,
            trans_primitives=[AddNumeric],
            max_depth=max_depth,
        )
        return dfs_obj

    for i in [-1, 0, 1, 2]:
        if i in [-1, 2]:
            match = (
                "Only one dataframe in entityset, changing max_depth to 1 "
                "since deeper features cannot be created"
            )
            with pytest.warns(UserWarning, match=match):
                dfs_obj = make_dfs_obj(i)
        else:
            dfs_obj = make_dfs_obj(i)

        features = dfs_obj.build_features()
        assert len(features) > 0
        if i != 0:
            # at least one depth 1 feature made
            assert any([f.get_depth() == 1 for f in features])
            # no depth 2 or higher even with max_depth=2
            assert all([f.get_depth() <= 1 for f in features])
        else:
            # no depth 1 or higher features with max_depth=0
            assert all([f.get_depth() == 0 for f in features])


def test_drop_contains(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Sum],
        trans_primitives=[],
        max_depth=1,
        seed_features=[],
        drop_contains=[],
    )
    features = dfs_obj.build_features()
    to_drop = features[2]
    partial_name = to_drop.get_name()[:5]

    dfs_drop = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Sum],
        trans_primitives=[],
        max_depth=1,
        seed_features=[],
        drop_contains=[partial_name],
    )
    features = dfs_drop.build_features()
    assert to_drop.get_name() not in [f.get_name() for f in features]


def test_drop_exact(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Sum],
        trans_primitives=[],
        max_depth=1,
        seed_features=[],
        drop_exact=[],
    )
    features = dfs_obj.build_features()
    to_drop = features[2]
    name = to_drop.get_name()
    dfs_drop = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Sum],
        trans_primitives=[],
        max_depth=1,
        seed_features=[],
        drop_exact=[name],
    )
    features = dfs_drop.build_features()
    assert name not in [f.get_name() for f in features]


def test_seed_features(es):
    seed_feature_sessions = (
        Feature(es["log"].ww["id"], parent_dataframe_name="sessions", primitive=Count)
        > 2
    )
    seed_feature_log = Feature(es["log"].ww["comments"], primitive=NumCharacters)
    session_agg = Feature(
        seed_feature_log,
        parent_dataframe_name="sessions",
        primitive=Mean,
    )
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Mean],
        trans_primitives=[],
        max_depth=2,
        seed_features=[seed_feature_sessions, seed_feature_log],
    )
    features = dfs_obj.build_features()
    assert seed_feature_sessions.get_name() in [f.get_name() for f in features]
    assert session_agg.get_name() in [f.get_name() for f in features]


def test_does_not_make_agg_of_direct_of_target_dataframe(es):
    # TODO: Update to work with Dask and Spark supported primitive
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Dask EntitySets do not support the Last primitive")

    count_sessions = Feature(
        es["sessions"].ww["id"],
        parent_dataframe_name="customers",
        primitive=Count,
    )
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[Last],
        trans_primitives=[],
        max_depth=2,
        seed_features=[count_sessions],
    )
    features = dfs_obj.build_features()
    # this feature is meaningless because customers.COUNT(sessions) is already defined on
    # the customers dataframe
    assert not feature_with_name(features, "LAST(sessions.customers.COUNT(sessions))")
    assert not feature_with_name(features, "LAST(sessions.customers.age)")


def test_dfs_builds_on_seed_features_more_than_max_depth(es):
    # TODO: Update to work with Dask and Spark supported primitive
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Dask EntitySets do not support the Last and Mode primitives")

    seed_feature_sessions = Feature(
        es["log"].ww["id"],
        parent_dataframe_name="sessions",
        primitive=Count,
    )
    seed_feature_log = Feature(es["log"].ww["datetime"], primitive=Hour)
    session_agg = Feature(
        seed_feature_log,
        parent_dataframe_name="sessions",
        primitive=Last,
    )

    # Depth of this feat is 2 relative to session_agg, the seed feature,
    # which is greater than max_depth so it shouldn't be built
    session_agg_trans = DirectFeature(
        Feature(session_agg, parent_dataframe_name="customers", primitive=Mode),
        "sessions",
    )
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Last, Count],
        trans_primitives=[],
        max_depth=1,
        seed_features=[seed_feature_sessions, seed_feature_log],
    )
    features = dfs_obj.build_features()
    assert seed_feature_sessions.get_name() in [f.get_name() for f in features]
    assert session_agg.get_name() in [f.get_name() for f in features]
    assert session_agg_trans.get_name() not in [f.get_name() for f in features]


def test_dfs_includes_seed_features_greater_than_max_depth(es):
    session_agg = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="sessions",
        primitive=Sum,
    )
    customer_agg = Feature(
        session_agg,
        parent_dataframe_name="customers",
        primitive=Mean,
    )
    assert customer_agg.get_depth() == 2

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[Mean],
        trans_primitives=[],
        max_depth=1,
        seed_features=[customer_agg],
    )
    features = dfs_obj.build_features()
    assert feature_with_name(features=features, name=customer_agg.get_name())


def test_allowed_paths(es):
    # TODO: Update to work with Dask and Spark supported primitive
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Dask EntitySets do not support the Last primitive")

    kwargs = dict(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[Last],
        trans_primitives=[],
        max_depth=2,
        seed_features=[],
    )
    dfs_unconstrained = DeepFeatureSynthesis(**kwargs)
    features_unconstrained = dfs_unconstrained.build_features()

    unconstrained_names = [f.get_name() for f in features_unconstrained]
    customers_session_feat = Feature(
        es["sessions"].ww["device_type"],
        parent_dataframe_name="customers",
        primitive=Last,
    )
    customers_session_log_feat = Feature(
        es["log"].ww["value"],
        parent_dataframe_name="customers",
        primitive=Last,
    )
    assert customers_session_feat.get_name() in unconstrained_names
    assert customers_session_log_feat.get_name() in unconstrained_names

    dfs_constrained = DeepFeatureSynthesis(
        allowed_paths=[["customers", "sessions"]], **kwargs
    )
    features = dfs_constrained.build_features()
    names = [f.get_name() for f in features]
    assert customers_session_feat.get_name() in names
    assert customers_session_log_feat.get_name() not in names


def test_max_features(es):
    kwargs = dict(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[Sum],
        trans_primitives=[],
        max_depth=2,
        seed_features=[],
    )
    dfs_unconstrained = DeepFeatureSynthesis(**kwargs)
    features_unconstrained = dfs_unconstrained.build_features()
    dfs_unconstrained_with_arg = DeepFeatureSynthesis(max_features=-1, **kwargs)
    feats_unconstrained_with_arg = dfs_unconstrained_with_arg.build_features()
    dfs_constrained = DeepFeatureSynthesis(max_features=1, **kwargs)
    features = dfs_constrained.build_features()
    assert len(features_unconstrained) == len(feats_unconstrained_with_arg)
    assert len(features) == 1


def test_where_primitives(es):
    es.add_interesting_values(dataframe_name="sessions", values={"device_type": [0]})
    kwargs = dict(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[Count, Sum],
        trans_primitives=[Absolute],
        max_depth=3,
    )
    dfs_unconstrained = DeepFeatureSynthesis(**kwargs)
    dfs_constrained = DeepFeatureSynthesis(where_primitives=["sum"], **kwargs)
    features_unconstrained = dfs_unconstrained.build_features()
    features = dfs_constrained.build_features()

    where_feats_unconstrained = [
        f
        for f in features_unconstrained
        if isinstance(f, AggregationFeature) and f.where is not None
    ]
    where_feats = [
        f for f in features if isinstance(f, AggregationFeature) and f.where is not None
    ]

    assert len(where_feats_unconstrained) >= 1

    assert (
        len([f for f in where_feats_unconstrained if isinstance(f.primitive, Sum)]) == 0
    )
    assert (
        len([f for f in where_feats_unconstrained if isinstance(f.primitive, Count)])
        > 0
    )

    assert len([f for f in where_feats if isinstance(f.primitive, Sum)]) > 0
    assert len([f for f in where_feats if isinstance(f.primitive, Count)]) == 0
    assert (
        len(
            [
                d
                for f in where_feats
                for d in f.get_dependencies(deep=True)
                if isinstance(d.primitive, Absolute)
            ],
        )
        > 0
    )


def test_stacking_where_primitives(es):
    # TODO: Update to work with Dask supported primitive
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Dask and Spark EntitySets do not support the Last primitive")
    es = copy.deepcopy(es)
    es.add_interesting_values(dataframe_name="sessions", values={"device_type": [0]})
    es.add_interesting_values(
        dataframe_name="log",
        values={"product_id": ["coke_zero"]},
    )
    kwargs = dict(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[Count, Last],
        max_depth=3,
    )
    dfs_where_stack_limit_1 = DeepFeatureSynthesis(
        where_primitives=["last", Count], **kwargs
    )
    dfs_where_stack_limit_2 = DeepFeatureSynthesis(
        where_primitives=["last", Count], where_stacking_limit=2, **kwargs
    )
    stack_limit_1_features = dfs_where_stack_limit_1.build_features()
    stack_limit_2_features = dfs_where_stack_limit_2.build_features()

    where_stack_1_feats = [
        f
        for f in stack_limit_1_features
        if isinstance(f, AggregationFeature) and f.where is not None
    ]
    where_stack_2_feats = [
        f
        for f in stack_limit_2_features
        if isinstance(f, AggregationFeature) and f.where is not None
    ]

    assert len(where_stack_1_feats) >= 1
    assert len(where_stack_2_feats) >= 1

    assert len([f for f in where_stack_1_feats if isinstance(f.primitive, Last)]) > 0
    assert len([f for f in where_stack_1_feats if isinstance(f.primitive, Count)]) > 0

    assert len([f for f in where_stack_2_feats if isinstance(f.primitive, Last)]) > 0
    assert len([f for f in where_stack_2_feats if isinstance(f.primitive, Count)]) > 0

    stacked_where_limit_1_feats = []
    stacked_where_limit_2_feats = []
    where_double_where_tuples = [
        (where_stack_1_feats, stacked_where_limit_1_feats),
        (where_stack_2_feats, stacked_where_limit_2_feats),
    ]
    for where_list, double_where_list in where_double_where_tuples:
        for feature in where_list:
            for base_feat in feature.base_features:
                if (
                    isinstance(base_feat, AggregationFeature)
                    and base_feat.where is not None
                ):
                    double_where_list.append(feature)

    assert len(stacked_where_limit_1_feats) == 0
    assert len(stacked_where_limit_2_feats) > 0


def test_where_different_base_feats(es):
    es.add_interesting_values(dataframe_name="sessions", values={"device_type": [0]})

    kwargs = dict(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[Sum, Count],
        where_primitives=[Sum, Count],
        max_depth=3,
    )
    dfs_unconstrained = DeepFeatureSynthesis(**kwargs)
    features = dfs_unconstrained.build_features()
    where_feats = [
        f.unique_name()
        for f in features
        if isinstance(f, AggregationFeature) and f.where is not None
    ]
    not_where_feats = [
        f.unique_name()
        for f in features
        if isinstance(f, AggregationFeature) and f.where is None
    ]
    for name in not_where_feats:
        assert name not in where_feats


def test_dfeats_where(es):
    # TODO: Update to work with Dask `es` fixture when issue #978 is closed
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Dask and Spark EntitySets do not support add_interesting_values")
    es.add_interesting_values()

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Count],
        trans_primitives=[],
    )

    features = dfs_obj.build_features()

    # test to make sure we build direct features of agg features with where clause
    assert feature_with_name(features, "customers.COUNT(log WHERE priority_level = 0)")

    assert feature_with_name(
        features,
        "COUNT(log WHERE products.department = electronics)",
    )


def test_commutative(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=es,
        agg_primitives=[Sum],
        trans_primitives=[AddNumeric],
        max_depth=3,
    )
    feats = dfs_obj.build_features()

    add_feats = [f for f in feats if isinstance(f.primitive, AddNumeric)]

    # Check that there are no two AddNumeric features with the same base
    # features.
    unordered_args = set()
    for f in add_feats:
        arg1, arg2 = f.base_features
        args_set = frozenset({arg1.unique_name(), arg2.unique_name()})
        unordered_args.add(args_set)

    assert len(add_feats) == len(unordered_args)


def test_transform_consistency(transform_es):
    # Generate features
    transform_es["first"].ww.set_types(
        logical_types={"b": "BooleanNullable", "b1": "BooleanNullable"},
    )
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="first",
        entityset=transform_es,
        trans_primitives=["and", "add_numeric", "or"],
        max_depth=1,
    )
    feature_defs = dfs_obj.build_features()

    # Check for correct ordering of features
    assert feature_with_name(feature_defs, "a")
    assert feature_with_name(feature_defs, "b")
    assert feature_with_name(feature_defs, "b1")
    assert feature_with_name(feature_defs, "b12")
    assert feature_with_name(feature_defs, "P")

    assert feature_with_name(feature_defs, "AND(b, b1)")
    assert not feature_with_name(
        feature_defs,
        "AND(b1, b)",
    )  # make sure it doesn't exist the other way
    assert feature_with_name(feature_defs, "a + P")
    assert feature_with_name(feature_defs, "b12 + P")
    assert feature_with_name(feature_defs, "a + b12")
    assert feature_with_name(feature_defs, "OR(b, b1)")


def test_transform_no_stack_agg(es):
    # TODO: Update to work with Dask and Spark supported primitives
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Dask EntitySets do not support the NMostCommon primitive")
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[NMostCommon],
        trans_primitives=[NotEqual],
        max_depth=3,
    )
    feature_defs = dfs_obj.build_features()

    assert not feature_with_name(
        feature_defs,
        "id != N_MOST_COMMON(sessions.device_type)",
    )


def test_initialized_trans_prim(es):
    prim = IsIn(list_of_outputs=["coke zero"])
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=es,
        agg_primitives=[],
        trans_primitives=[prim],
    )

    features = dfs_obj.build_features()

    assert feature_with_name(features, "product_id.isin(['coke zero'])")


def test_initialized_agg_prim(es):
    # TODO: Update to work with Dask and Spark supported primitives
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail("Dask EntitySets do not support the NMostCommon primitive")
    ThreeMost = NMostCommon(n=3)
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[ThreeMost],
        trans_primitives=[],
    )
    features = dfs_obj.build_features()

    assert feature_with_name(features, "N_MOST_COMMON(log.subregioncode)")


def test_return_types(es):
    # TODO: Update to work with Dask and Spark supported primitive
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail(
            "Dask and Spark EntitySets do not support the NMostCommon primitive",
        )
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[Count, NMostCommon],
        trans_primitives=[Absolute, Hour, IsIn],
    )

    discrete = ColumnSchema(semantic_tags={"category"})
    numeric = ColumnSchema(semantic_tags={"numeric"})
    datetime = ColumnSchema(logical_type=Datetime)

    f1 = dfs_obj.build_features(return_types=None)
    f2 = dfs_obj.build_features(return_types=[discrete])
    f3 = dfs_obj.build_features(return_types="all")
    f4 = dfs_obj.build_features(return_types=[datetime])

    f1_types = [f.column_schema for f in f1]
    f2_types = [f.column_schema for f in f2]
    f3_types = [f.column_schema for f in f3]
    f4_types = [f.column_schema for f in f4]

    assert any([is_valid_input(schema, discrete) for schema in f1_types])
    assert any([is_valid_input(schema, numeric) for schema in f1_types])
    assert not any([is_valid_input(schema, datetime) for schema in f1_types])

    assert any([is_valid_input(schema, discrete) for schema in f2_types])
    assert not any([is_valid_input(schema, numeric) for schema in f2_types])
    assert not any([is_valid_input(schema, datetime) for schema in f2_types])

    assert any([is_valid_input(schema, discrete) for schema in f3_types])
    assert any([is_valid_input(schema, numeric) for schema in f3_types])
    assert any([is_valid_input(schema, datetime) for schema in f3_types])

    assert not any([is_valid_input(schema, discrete) for schema in f4_types])
    assert not any([is_valid_input(schema, numeric) for schema in f4_types])
    assert any([is_valid_input(schema, datetime) for schema in f4_types])


def test_checks_primitives_correct_type(es):
    error_text = (
        "Primitive <class \\'featuretools\\.primitives\\.standard\\."
        "transform\\.datetime\\.hour\\.Hour\\'> in "
        "agg_primitives is not an aggregation primitive"
    )
    with pytest.raises(ValueError, match=error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="sessions",
            entityset=es,
            agg_primitives=[Hour],
            trans_primitives=[],
        )

    error_text = (
        "Primitive <class \\'featuretools\\.primitives\\.standard\\."
        "aggregation\\.sum_primitive\\.Sum\\'> in trans_primitives "
        "is not a transform primitive"
    )
    with pytest.raises(ValueError, match=error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="sessions",
            entityset=es,
            agg_primitives=[],
            trans_primitives=[Sum],
        )


def test_makes_agg_features_along_multiple_paths(diamond_es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="regions",
        entityset=diamond_es,
        agg_primitives=["mean"],
        trans_primitives=[],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "MEAN(customers.transactions.amount)")
    assert feature_with_name(features, "MEAN(stores.transactions.amount)")


def test_makes_direct_features_through_multiple_relationships(games_es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="games",
        entityset=games_es,
        agg_primitives=["mean"],
        trans_primitives=[],
    )

    features = dfs_obj.build_features()

    teams = ["home", "away"]
    for forward in teams:
        for backward in teams:
            for col in teams:
                f = "teams[%s_team_id].MEAN(games[%s_team_id].%s_team_score)" % (
                    forward,
                    backward,
                    col,
                )
                assert feature_with_name(features, f)


def test_stacks_multioutput_features(es):
    # TODO: Update to work with Dask and Spark supported primitive
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail(
            "Dask EntitySets do not support the NumUnique and NMostCommon primitives",
        )

    class TestTime(TransformPrimitive):
        name = "test_time"
        input_types = [ColumnSchema(logical_type=Datetime)]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        number_output_features = 6

        def get_function(self):
            def test_f(x):
                times = pd.Series(x)
                units = ["year", "month", "day", "hour", "minute", "second"]
                return [times.apply(lambda x: getattr(x, unit)) for unit in units]

            return test_f

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[NumUnique, NMostCommon(n=3)],
        trans_primitives=[TestTime, Diff],
        max_depth=4,
    )
    feat = dfs_obj.build_features()

    for i in range(3):
        f = "NUM_UNIQUE(sessions.N_MOST_COMMON(log.countrycode)[%d])" % i
        assert feature_with_name(feat, f)


def test_seed_multi_output_feature_stacking(es):
    # TODO: Update to work with Dask and Spark supported primitive
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail(
            "Dask EntitySets do not support the NMostCommon and NumUnique primitives",
        )
    threecommon = NMostCommon(3)
    tc = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="sessions",
        primitive=threecommon,
    )

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=es,
        seed_features=[tc],
        agg_primitives=[NumUnique],
        trans_primitives=[],
        max_depth=4,
    )
    feat = dfs_obj.build_features()

    for i in range(3):
        f = "NUM_UNIQUE(sessions.N_MOST_COMMON(log.product_id)[%d])" % i
        assert feature_with_name(feat, f)


def test_makes_direct_features_along_multiple_paths(diamond_es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="transactions",
        entityset=diamond_es,
        max_depth=3,
        agg_primitives=[],
        trans_primitives=[],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "customers.regions.name")
    assert feature_with_name(features, "stores.regions.name")


def test_does_not_make_trans_of_single_direct_feature(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[],
        trans_primitives=["weekday"],
        max_depth=2,
    )

    features = dfs_obj.build_features()

    assert not feature_with_name(features, "WEEKDAY(customers.signup_date)")
    assert feature_with_name(features, "customers.WEEKDAY(signup_date)")


def test_makes_trans_of_multiple_direct_features(diamond_es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="transactions",
        entityset=diamond_es,
        agg_primitives=["mean"],
        trans_primitives=[Equal],
        max_depth=4,
    )

    features = dfs_obj.build_features()

    # Make trans of direct and non-direct
    assert feature_with_name(features, "amount = stores.MEAN(transactions.amount)")

    # Make trans of direct features on different dataframes
    assert feature_with_name(
        features,
        "customers.MEAN(transactions.amount) = stores.square_ft",
    )

    # Make trans of direct features on same dataframe with different paths.
    assert feature_with_name(features, "customers.regions.name = stores.regions.name")

    # Don't make trans of direct features with same path.
    assert not feature_with_name(
        features,
        "stores.square_ft = stores.MEAN(transactions.amount)",
    )
    assert not feature_with_name(
        features,
        "stores.MEAN(transactions.amount) = stores.square_ft",
    )

    # The naming of the below is confusing but this is a direct feature of a transform.
    assert feature_with_name(features, "stores.MEAN(transactions.amount) = square_ft")


def test_makes_direct_of_agg_of_trans_on_target(es):
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=es,
        agg_primitives=["mean"],
        trans_primitives=[Absolute],
        max_depth=3,
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "sessions.MEAN(log.ABSOLUTE(value))")


def test_primitive_options_errors(es):
    wrong_key_options = {"mean": {"ignore_dataframe": ["sessions"]}}
    wrong_type_list = {"mean": {"ignore_dataframes": "sessions"}}
    wrong_type_dict = {"mean": {"ignore_columns": {"sessions": "product_id"}}}
    conflicting_primitive_options = {
        ("count", "mean"): {"ignore_dataframes": ["sessions"]},
        "mean": {"include_dataframes": ["sessions"]},
    }
    invalid_dataframe = {"mean": {"include_dataframes": ["invalid_dataframe"]}}
    invalid_column_dataframe = {
        "mean": {"include_columns": {"invalid_dataframe": ["product_id"]}},
    }
    invalid_column = {"mean": {"include_columns": {"sessions": ["invalid_column"]}}}
    key_error_text = "Unrecognized primitive option 'ignore_dataframe' for mean"
    list_error_text = "Incorrect type formatting for 'ignore_dataframes' for mean"
    dict_error_text = "Incorrect type formatting for 'ignore_columns' for mean"
    conflicting_error_text = "Multiple options found for primitive mean"
    invalid_dataframe_warning = "Dataframe 'invalid_dataframe' not in entityset"
    invalid_column_warning = "Column 'invalid_column' not in dataframe 'sessions'"
    with pytest.raises(KeyError, match=key_error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=["mean"],
            trans_primitives=[],
            primitive_options=wrong_key_options,
        )
    with pytest.raises(TypeError, match=list_error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=["mean"],
            trans_primitives=[],
            primitive_options=wrong_type_list,
        )
    with pytest.raises(TypeError, match=dict_error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=["mean"],
            trans_primitives=[],
            primitive_options=wrong_type_dict,
        )
    with pytest.raises(KeyError, match=conflicting_error_text):
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=["mean"],
            trans_primitives=[],
            primitive_options=conflicting_primitive_options,
        )
    with pytest.warns(UserWarning, match=invalid_dataframe_warning) as record:
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=["mean"],
            trans_primitives=[],
            primitive_options=invalid_dataframe,
        )
    assert len(record) == 1
    with pytest.warns(UserWarning, match=invalid_dataframe_warning) as record:
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=["mean"],
            trans_primitives=[],
            primitive_options=invalid_column_dataframe,
        )
    assert len(record) == 1
    with pytest.warns(UserWarning, match=invalid_column_warning) as record:
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=["mean"],
            trans_primitives=[],
            primitive_options=invalid_column,
        )
    assert len(record) == 1


def test_primitive_options(es):
    options = {
        "sum": {"include_columns": {"customers": ["age"]}},
        "mean": {"include_dataframes": ["customers"]},
        "mode": {"ignore_dataframes": ["sessions"]},
        "num_unique": {"ignore_columns": {"customers": ["engagement_level"]}},
    }
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="cohorts",
        entityset=es,
        primitive_options=options,
    )
    features = dfs_obj.build_features()

    for f in features:
        deps = f.get_dependencies(deep=True)
        df_names = [d.dataframe_name for d in deps]
        columns = [d for d in deps if isinstance(d, IdentityFeature)]
        if isinstance(f.primitive, Sum):
            for identity_base in columns:
                if identity_base.dataframe_name == "customers":
                    assert identity_base.get_name() == "age"
        if isinstance(f.primitive, Mean):
            assert all([df_name in ["customers"] for df_name in df_names])
        if isinstance(f.primitive, Mode):
            assert "sessions" not in df_names
        if isinstance(f.primitive, NumUnique):
            for identity_base in columns:
                assert not (
                    identity_base.dataframe_name == "customers"
                    and identity_base.get_name() == "engagement_level"
                )

    options = {
        "month": {"ignore_columns": {"customers": ["birthday"]}},
        "day": {"include_columns": {"customers": ["signup_date", "upgrade_date"]}},
        "num_characters": {"ignore_dataframes": ["customers"]},
        "year": {"include_dataframes": ["customers"]},
    }
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=es,
        agg_primitives=[],
        ignore_dataframes=["cohort"],
        primitive_options=options,
    )
    features = dfs_obj.build_features()
    assert not any([isinstance(f, NumCharacters) for f in features])
    for f in features:
        deps = f.get_dependencies(deep=True)
        df_names = [d.dataframe_name for d in deps]
        columns = [d for d in deps if isinstance(d, IdentityFeature)]
        if isinstance(f.primitive, Month):
            for identity_base in columns:
                assert not (
                    identity_base.dataframe_name == "customers"
                    and identity_base.get_name() == "birthday"
                )
        if isinstance(f.primitive, Day):
            for identity_base in columns:
                if identity_base.dataframe_name == "customers":
                    assert (
                        identity_base.get_name() == "signup_date"
                        or identity_base.get_name() == "upgrade_date"
                    )
        if isinstance(f.primitive, Year):
            assert all([df_name in ["customers"] for df_name in df_names])


def test_primitive_options_with_globals(es):
    # non-overlapping ignore_dataframes
    options = {"mode": {"ignore_dataframes": ["sessions"]}}
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="cohorts",
        entityset=es,
        ignore_dataframes=["régions"],
        primitive_options=options,
    )
    features = dfs_obj.build_features()
    for f in features:
        deps = f.get_dependencies(deep=True)
        df_names = [d.dataframe_name for d in deps]
        assert "régions" not in df_names
        if isinstance(f.primitive, Mode):
            assert "sessions" not in df_names

    # non-overlapping ignore_columns
    options = {"num_unique": {"ignore_columns": {"customers": ["engagement_level"]}}}
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=es,
        ignore_columns={"customers": ["région_id"]},
        primitive_options=options,
    )
    features = dfs_obj.build_features()
    for f in features:
        deps = f.get_dependencies(deep=True)
        columns = [d for d in deps if isinstance(d, IdentityFeature)]
        for identity_base in columns:
            assert not (
                identity_base.dataframe_name == "customers"
                and identity_base.get_name() == "région_id"
            )
        if isinstance(f.primitive, NumUnique):
            for identity_base in columns:
                assert not (
                    identity_base.dataframe_name == "customers"
                    and identity_base.get_name() == "engagement_level"
                )

    # Overlapping globals/options with ignore_dataframes
    options = {
        "mode": {
            "include_dataframes": ["sessions", "customers"],
            "ignore_columns": {"customers": ["région_id"]},
        },
        "num_unique": {
            "include_dataframes": ["sessions", "customers"],
            "include_columns": {"sessions": ["device_type"], "customers": ["age"]},
        },
        "month": {"ignore_columns": {"cohorts": ["cohort_end"]}},
    }
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="cohorts",
        entityset=es,
        ignore_dataframes=["sessions"],
        ignore_columns={"customers": ["age"]},
        primitive_options=options,
    )
    features = dfs_obj.build_features()
    for f in features:
        assert f.primitive.name != "month"
        # ignoring cohorts means no features are created
        assert not isinstance(f.primitive, Month)

        deps = f.get_dependencies(deep=True)
        df_names = [d.dataframe_name for d in deps]
        columns = [d for d in deps if isinstance(d, IdentityFeature)]
        if isinstance(f.primitive, Mode):
            assert [all([df_name in ["sessions", "customers"] for df_name in df_names])]
            for identity_base in columns:
                assert not (
                    identity_base.dataframe_name == "customers"
                    and (
                        identity_base.get_name() == "age"
                        or identity_base.get_name() == "région_id"
                    )
                )
        elif isinstance(f.primitive, NumUnique):
            assert [all([df_name in ["sessions", "customers"] for df_name in df_names])]
            for identity_base in columns:
                if identity_base.dataframe_name == "sessions":
                    assert identity_base.get_name() == "device_type"
        # All other primitives ignore 'sessions' and 'age'
        else:
            assert "sessions" not in df_names
            for identity_base in columns:
                assert not (
                    identity_base.dataframe_name == "customers"
                    and identity_base.get_name() == "age"
                )


def test_primitive_options_groupbys(pd_es):
    options = {
        "cum_count": {"include_groupby_dataframes": ["log", "customers"]},
        "cum_sum": {"ignore_groupby_dataframes": ["sessions"]},
        "cum_mean": {
            "ignore_groupby_columns": {
                "customers": ["région_id"],
                "log": ["session_id"],
            },
        },
        "cum_min": {
            "include_groupby_columns": {"sessions": ["customer_id", "device_type"]},
        },
    }

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[],
        trans_primitives=[],
        max_depth=3,
        groupby_trans_primitives=["cum_sum", "cum_count", "cum_min", "cum_mean"],
        primitive_options=options,
    )
    features = dfs_obj.build_features()
    for f in features:
        if isinstance(f, GroupByTransformFeature):
            deps = f.groupby.get_dependencies(deep=True)
            df_names = [d.dataframe_name for d in deps] + [f.groupby.dataframe_name]
            columns = [d for d in deps if isinstance(d, IdentityFeature)]
            columns += [f.groupby] if isinstance(f.groupby, IdentityFeature) else []
        if isinstance(f.primitive, CumMean):
            for identity_groupby in columns:
                assert not (
                    identity_groupby.dataframe_name == "customers"
                    and identity_groupby.get_name() == "région_id"
                )
                assert not (
                    identity_groupby.dataframe_name == "log"
                    and identity_groupby.get_name() == "session_id"
                )
        if isinstance(f.primitive, CumCount):
            assert all([name in ["log", "customers"] for name in df_names])
        if isinstance(f.primitive, CumSum):
            assert "sessions" not in df_names
        if isinstance(f.primitive, CumMin):
            for identity_groupby in columns:
                if identity_groupby.dataframe_name == "sessions":
                    assert (
                        identity_groupby.get_name() == "customer_id"
                        or identity_groupby.get_name() == "device_type"
                    )


def test_primitive_options_multiple_inputs(es):
    if es.dataframe_type != Library.PANDAS:
        pytest.xfail(
            "Dask and Spark EntitySets do not support various primitives used in this test",
        )
    too_many_options = {
        "mode": [{"include_dataframes": ["logs"]}, {"ignore_dataframes": ["sessions"]}],
    }
    error_msg = "Number of options does not match number of inputs for primitive mode"
    with pytest.raises(AssertionError, match=error_msg):
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=["mode"],
            trans_primitives=[],
            primitive_options=too_many_options,
        )

    unknown_primitive = Trend()
    unknown_primitive.name = "unknown_primitive"
    unknown_primitive_option = {
        "unknown_primitive": [
            {"include_dataframes": ["logs"]},
            {"ignore_dataframes": ["sessions"]},
        ],
    }
    error_msg = "Unknown primitive with name 'unknown_primitive'"
    with pytest.raises(ValueError, match=error_msg):
        DeepFeatureSynthesis(
            target_dataframe_name="customers",
            entityset=es,
            agg_primitives=[unknown_primitive],
            trans_primitives=[],
            primitive_options=unknown_primitive_option,
        )

    options1 = {
        "trend": [
            {"include_dataframes": ["log"], "ignore_columns": {"log": ["value"]}},
            {"include_dataframes": ["log"], "include_columns": {"log": ["datetime"]}},
        ],
    }
    dfs_obj1 = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=["trend"],
        trans_primitives=[],
        primitive_options=options1,
    )
    features1 = dfs_obj1.build_features()
    for f in features1:
        deps = f.get_dependencies()
        df_names = [d.dataframe_name for d in deps]
        columns = [d.get_name() for d in deps]
        if f.primitive.name == "trend":
            assert all([df_name in ["log"] for df_name in df_names])
            assert "datetime" in columns
            if len(columns) == 2:
                assert "value" != columns[0]

    options2 = {
        Trend: [
            {"include_dataframes": ["log"], "ignore_columns": {"log": ["value"]}},
            {"include_dataframes": ["log"], "include_columns": {"log": ["datetime"]}},
        ],
    }
    dfs_obj2 = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=["trend"],
        trans_primitives=[],
        primitive_options=options2,
    )
    features2 = dfs_obj2.build_features()

    assert set(features2) == set(features1)


def test_primitive_options_class_names(es):
    options1 = {"mean": {"include_dataframes": ["customers"]}}

    options2 = {Mean: {"include_dataframes": ["customers"]}}

    bad_options = {
        "mean": {"include_dataframes": ["customers"]},
        Mean: {"ignore_dataframes": ["customers"]},
    }
    conflicting_error_text = "Multiple options found for primitive mean"

    primitives = [["mean"], [Mean]]
    options = [options1, options2]

    features = []
    for primitive in primitives:
        with pytest.raises(KeyError, match=conflicting_error_text):
            DeepFeatureSynthesis(
                target_dataframe_name="cohorts",
                entityset=es,
                agg_primitives=primitive,
                trans_primitives=[],
                primitive_options=bad_options,
            )
        for option in options:
            dfs_obj = DeepFeatureSynthesis(
                target_dataframe_name="cohorts",
                entityset=es,
                agg_primitives=primitive,
                trans_primitives=[],
                primitive_options=option,
            )
            features.append(set(dfs_obj.build_features()))

    for f in features[0]:
        deps = f.get_dependencies(deep=True)
        df_names = [d.dataframe_name for d in deps]
        if isinstance(f.primitive, Mean):
            assert all(df_name == "customers" for df_name in df_names)

    assert features[0] == features[1] == features[2] == features[3]


def test_primitive_options_instantiated_primitive(es):
    warning_msg = (
        "Options present for primitive instance and generic "
        "primitive class \\(mean\\), primitive instance will not use generic "
        "options"
    )

    skipna_mean = Mean(skipna=False)
    options = {
        skipna_mean: {"include_dataframes": ["stores"]},
        "mean": {"ignore_dataframes": ["stores"]},
    }
    with pytest.warns(UserWarning, match=warning_msg):
        dfs_obj = DeepFeatureSynthesis(
            target_dataframe_name="régions",
            entityset=es,
            agg_primitives=["mean", skipna_mean],
            trans_primitives=[],
            primitive_options=options,
        )

    features = dfs_obj.build_features()
    for f in features:
        deps = f.get_dependencies(deep=True)
        df_names = [d.dataframe_name for d in deps]
        if f.primitive == skipna_mean:
            assert all(df_name == "stores" for df_name in df_names)
        elif isinstance(f.primitive, Mean):
            assert "stores" not in df_names


def test_primitive_options_commutative(es):
    class AddThree(TransformPrimitive):
        name = "add_three"
        input_types = [
            ColumnSchema(semantic_tags={"numeric"}),
            ColumnSchema(semantic_tags={"numeric"}),
            ColumnSchema(semantic_tags={"numeric"}),
        ]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        commutative = True
        compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]

        def generate_name(self, base_feature_names):
            return "%s + %s + %s" % (
                base_feature_names[0],
                base_feature_names[1],
                base_feature_names[2],
            )

    options = {
        "add_numeric": [
            {"include_columns": {"log": ["value_2"]}},
            {"include_columns": {"log": ["value"]}},
        ],
        AddThree: [
            {"include_columns": {"log": ["value_2"]}},
            {"include_columns": {"log": ["value_many_nans"]}},
            {"include_columns": {"log": ["value"]}},
        ],
    }
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=es,
        agg_primitives=[],
        trans_primitives=[AddNumeric, AddThree],
        primitive_options=options,
        max_depth=1,
    )
    features = dfs_obj.build_features()
    add_numeric = [f for f in features if isinstance(f.primitive, AddNumeric)]
    assert len(add_numeric) == 1
    deps = add_numeric[0].get_dependencies(deep=True)
    assert deps[0].get_name() == "value_2" and deps[1].get_name() == "value"

    add_three = [f for f in features if isinstance(f.primitive, AddThree)]
    assert len(add_three) == 1
    deps = add_three[0].get_dependencies(deep=True)
    assert (
        deps[0].get_name() == "value_2"
        and deps[1].get_name() == "value_many_nans"
        and deps[2].get_name() == "value"
    )


def test_primitive_options_include_over_exclude(es):
    options = {
        "mean": {"ignore_dataframes": ["stores"], "include_dataframes": ["stores"]},
    }
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="régions",
        entityset=es,
        agg_primitives=["mean"],
        trans_primitives=[],
        primitive_options=options,
    )

    features = dfs_obj.build_features()
    at_least_one_mean = False
    for f in features:
        deps = f.get_dependencies(deep=True)
        dataframes = [d.dataframe_name for d in deps]
        if isinstance(f.primitive, Mean):
            at_least_one_mean = True
            assert "stores" in dataframes
    assert at_least_one_mean


def test_primitive_ordering():
    # Test that the order of the input primitives impacts neither
    # which features are created nor their order
    es = make_ecommerce_entityset()

    trans_prims = [AddNumeric, Absolute, "divide_numeric", NotEqual, "is_null"]
    groupby_trans_prim = ["cum_mean", CumMin, CumSum]
    agg_prims = [NMostCommon(n=3), Sum, Mean, Mean(skipna=False), "min", "max"]
    where_prims = ["count", Sum]

    seed_num_chars = Feature(
        es["customers"].ww["favorite_quote"],
        primitive=NumCharacters,
    )
    seed_is_null = Feature(es["customers"].ww["age"], primitive=IsNull)
    seed_features = [seed_num_chars, seed_is_null]

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=es,
        trans_primitives=trans_prims,
        groupby_trans_primitives=groupby_trans_prim,
        agg_primitives=agg_prims,
        where_primitives=where_prims,
        seed_features=seed_features,
        max_features=-1,
        max_depth=2,
    )
    features1 = dfs_obj.build_features()

    trans_prims.reverse()
    groupby_trans_prim.reverse()
    agg_prims.reverse()
    where_prims.reverse()
    seed_features.reverse()

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="customers",
        entityset=es,
        trans_primitives=trans_prims,
        groupby_trans_primitives=groupby_trans_prim,
        agg_primitives=agg_prims,
        where_primitives=where_prims,
        seed_features=seed_features,
        max_features=-1,
        max_depth=2,
    )
    features2 = dfs_obj.build_features()

    assert len(features1) == len(features2)

    for i in range(len(features2)):
        assert features1[i].unique_name() == features2[i].unique_name()


def test_no_transform_stacking():
    df1 = pd.DataFrame({"id": [0, 1, 2, 3], "A": [0, 1, 2, 3]})
    df2 = pd.DataFrame(
        {"index": [0, 1, 2, 3], "first_id": [0, 1, 1, 3], "B": [99, 88, 77, 66]},
    )

    dataframes = {"first": (df1, "id"), "second": (df2, "index")}
    relationships = [("first", "id", "second", "first_id")]
    es = EntitySet("data", dataframes, relationships)

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="second",
        entityset=es,
        trans_primitives=["negate", "add_numeric"],
        agg_primitives=["sum"],
        max_depth=4,
    )
    feature_defs = dfs_obj.build_features()

    expected = [
        "first_id",
        "B",
        "-(B)",
        "first.A",
        "first.SUM(second.B)",
        "first.-(A)",
        "B + first.A",
        "first.SUM(second.-(B))",
        "first.A + SUM(second.B)",
        "first.-(SUM(second.B))",
        "B + first.SUM(second.B)",
        "first.A + SUM(second.-(B))",
        "first.SUM(second.-(B)) + SUM(second.B)",
        "first.-(SUM(second.-(B)))",
        "B + first.SUM(second.-(B))",
    ]

    assert len(feature_defs) == len(expected)

    for feature_name in expected:
        assert feature_with_name(feature_defs, feature_name)


def test_builds_seed_features_on_foreign_key_col(es):
    seed_feature_sessions = Feature(es["sessions"].ww["customer_id"], primitive=Negate)

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        agg_primitives=[],
        trans_primitives=[],
        max_depth=2,
        seed_features=[seed_feature_sessions],
    )

    features = dfs_obj.build_features()
    assert feature_with_name(features, "-(customer_id)")


def test_does_not_build_features_on_last_time_index_col(es):
    es.add_last_time_indexes()

    dfs_obj = DeepFeatureSynthesis(target_dataframe_name="log", entityset=es)

    features = dfs_obj.build_features()

    for feature in features:
        assert LTI_COLUMN_NAME not in feature.get_name()


def test_builds_features_using_all_input_types(es):
    if es.dataframe_type == Library.SPARK:
        pytest.skip("NumTrue primitive not compatible with Spark")
    new_log_df = es["log"]
    new_log_df.ww["purchased_nullable"] = es["log"]["purchased"]
    new_log_df.ww.set_types(logical_types={"purchased_nullable": "boolean_nullable"})
    es.replace_dataframe("log", new_log_df)

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=es,
        trans_primitives=[Not],
        max_depth=1,
    )
    trans_features = dfs_obj.build_features()
    assert feature_with_name(trans_features, "NOT(purchased)")
    assert feature_with_name(trans_features, "NOT(purchased_nullable)")

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=es,
        groupby_trans_primitives=[Not],
        max_depth=1,
    )
    groupby_trans_features = dfs_obj.build_features()
    assert feature_with_name(groupby_trans_features, "NOT(purchased) by session_id")
    assert feature_with_name(
        groupby_trans_features,
        "NOT(purchased_nullable) by session_id",
    )

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="sessions",
        entityset=es,
        trans_primitives=[],
        agg_primitives=[NumTrue],
    )
    agg_features = dfs_obj.build_features()
    assert feature_with_name(agg_features, "NUM_TRUE(log.purchased)")
    assert feature_with_name(agg_features, "NUM_TRUE(log.purchased_nullable)")


def test_make_groupby_features_with_depth_none(pd_es):
    # If max_depth is set to -1, it sets it to None internally, so this
    # test validates code paths that have a None max_depth
    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[],
        trans_primitives=[],
        groupby_trans_primitives=["cum_sum"],
        max_depth=-1,
    )
    features = dfs_obj.build_features()
    assert feature_with_name(features, "CUM_SUM(value) by session_id")


def test_check_stacking_when_building_transform_features(pd_es):
    class NewMean(Mean):
        name = "NEW_MEAN"
        base_of_exclude = [Absolute]

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[NewMean, "mean"],
        trans_primitives=["absolute"],
        max_depth=-1,
    )
    features = dfs_obj.build_features()
    assert number_of_features_with_name_like(features, "ABSOLUTE(MEAN") > 0
    assert number_of_features_with_name_like(features, "ABSOLUTE(NEW_MEAN") == 0


def test_check_stacking_when_building_groupby_features(pd_es):
    class NewMean(Mean):
        name = "NEW_MEAN"
        base_of_exclude = [CumSum]

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=[NewMean, "mean"],
        groupby_trans_primitives=["cum_sum"],
        max_depth=5,
    )
    features = dfs_obj.build_features()
    assert number_of_features_with_name_like(features, "CUM_SUM(MEAN") > 0
    assert number_of_features_with_name_like(features, "CUM_SUM(NEW_MEAN") == 0


def test_check_stacking_when_building_agg_features(pd_es):
    class NewAbsolute(Absolute):
        name = "NEW_ABSOLUTE"
        base_of_exclude = [Mean]

    dfs_obj = DeepFeatureSynthesis(
        target_dataframe_name="log",
        entityset=pd_es,
        agg_primitives=["mean"],
        trans_primitives=[NewAbsolute, "absolute"],
        max_depth=5,
    )
    features = dfs_obj.build_features()
    assert number_of_features_with_name_like(features, "MEAN(log.ABSOLUTE") > 0
    assert number_of_features_with_name_like(features, "MEAN(log.NEW_ABSOLUTE") == 0



from datetime import datetime

import numpy as np
import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Double

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils import convert_time_units
from featuretools.utils.gen_utils import Library


class AvgTimeBetween(AggregationPrimitive):
    """Computes the average number of seconds between consecutive events.

    Description:
        Given a list of datetimes, return the average time (default in seconds)
        elapsed between consecutive events. If there are fewer
        than 2 non-null values, return `NaN`.

    Args:
        unit (str): Defines the unit of time.
            Defaults to seconds. Acceptable values:
            years, months, days, hours, minutes, seconds, milliseconds, nanoseconds

    Examples:
        >>> from datetime import datetime
        >>> avg_time_between = AvgTimeBetween()
        >>> times = [datetime(2010, 1, 1, 11, 45, 0),
        ...          datetime(2010, 1, 1, 11, 55, 15),
        ...          datetime(2010, 1, 1, 11, 57, 30)]
        >>> avg_time_between(times)
        375.0
        >>> avg_time_between = AvgTimeBetween(unit="minutes")
        >>> avg_time_between(times)
        6.25
    """

    name = "avg_time_between"
    input_types = [ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"})]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})
    description_template = "the average time between each of {}"

    def __init__(self, unit="seconds"):
        self.unit = unit.lower()

    def get_function(self, agg_type=Library.PANDAS):
        def pd_avg_time_between(x):
            """Assumes time scales are closer to order
            of seconds than to nanoseconds
            if times are much closer to nanoseconds
            we could get some floating point errors

            this can be fixed with another function
            that calculates the mean before converting
            to seconds
            """
            x = x.dropna()
            if x.shape[0] < 2:
                return np.nan
            if isinstance(x.iloc[0], (pd.Timestamp, datetime)):
                x = x.view("int64")
                # use len(x)-1 because we care about difference
                # between values, len(x)-1 = len(diff(x))

            avg = (x.max() - x.min()) / (len(x) - 1)
            avg = avg * 1e-9

            # long form:
            # diff_in_ns = x.diff().iloc[1:].astype('int64')
            # diff_in_seconds = diff_in_ns * 1e-9
            # avg = diff_in_seconds.mean()
            return convert_time_units(avg, self.unit)

        return pd_avg_time_between



from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base import TransformPrimitive


class Diff(TransformPrimitive):
    """Computes the difference between the value in a list and the
    previous value in that list.

    Args:
        periods (int): The number of periods by which to shift the index row.
            Default is 0. Periods correspond to rows.

    Description:
        Given a list of values, compute the difference from the previous
        item in the list. The result for the first element of the list will
        always be `NaN`.

    Examples:
        >>> diff = Diff()
        >>> values = [1, 10, 3, 4, 15]
        >>> diff(values).tolist()
        [nan, 9.0, -7.0, 1.0, 11.0]

        You can specify the number of periods to shift the values

        >>> values = [1, 2, 4, 7, 11, 16]
        >>> diff_periods = Diff(periods = 1)
        >>> diff_periods(values).tolist()
        [nan, nan, 1.0, 2.0, 3.0, 4.0]
    """

    name = "diff"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    uses_full_dataframe = True
    description_template = "the difference from the previous value of {}"

    def __init__(self, periods=0):
        self.periods = periods

    def get_function(self):
        def pd_diff(values):
            return values.shift(self.periods).diff()

        return pd_diff



import pandas as pd
from woodwork.logical_types import NaturalLanguage

import featuretools as ft


def load_retail(id="demo_retail_data", nrows=None, return_single_table=False):
    """Returns the retail entityset example.
    The original dataset can be found `here <https://archive.ics.uci.edu/ml/datasets/online+retail>`_.

    We have also made some modifications to the data. We
    changed the column names, converted the ``customer_id``
    to a unique fake ``customer_name``, dropped duplicates,
    added columns for ``total`` and ``cancelled`` and
    converted amounts from GBP to USD. You can download the modified CSV in gz `compressed (7 MB)
    <https://oss.alteryx.com/datasets/online-retail-logs-2018-08-28.csv.gz>`_
    or `uncompressed (43 MB)
    <https://oss.alteryx.com/datasets/online-retail-logs-2018-08-28.csv>`_ formats.

    Args:
        id (str):  Id to assign to EntitySet.
        nrows (int):  Number of rows to load of the underlying CSV.
            If None, load all.
        return_single_table (bool): If True, return a CSV rather than an EntitySet. Default is False.

    Examples:

        .. ipython::
            :verbatim:

            In [1]: import featuretools as ft

            In [2]: es = ft.demo.load_retail()

            In [3]: es
            Out[3]:
            Entityset: demo_retail_data
              DataFrames:
                orders (shape = [22190, 3])
                products (shape = [3684, 3])
                customers (shape = [4372, 2])
                order_products (shape = [401704, 7])

        Load in subset of data

        .. ipython::
            :verbatim:

            In [4]: es = ft.demo.load_retail(nrows=1000)

            In [5]: es
            Out[5]:
            Entityset: demo_retail_data
              DataFrames:
                orders (shape = [67, 5])
                products (shape = [606, 3])
                customers (shape = [50, 2])
                order_products (shape = [1000, 7])
    """
    es = ft.EntitySet(id)
    csv_s3_gz = (
        "https://oss.alteryx.com/datasets/online-retail-logs-2018-08-28.csv.gz?library=featuretools&version="
        + ft.__version__
    )
    csv_s3 = (
        "https://oss.alteryx.com/datasets/online-retail-logs-2018-08-28.csv?library=featuretools&version="
        + ft.__version__
    )
    # Try to read in gz compressed file
    try:
        df = pd.read_csv(csv_s3_gz, nrows=nrows, parse_dates=["order_date"])
    # Fall back to uncompressed
    except Exception:
        df = pd.read_csv(csv_s3, nrows=nrows, parse_dates=["order_date"])
    if return_single_table:
        return df

    es.add_dataframe(
        dataframe_name="order_products",
        dataframe=df,
        index="order_product_id",
        make_index=True,
        time_index="order_date",
        logical_types={"description": NaturalLanguage},
    )

    es.normalize_dataframe(
        new_dataframe_name="products",
        base_dataframe_name="order_products",
        index="product_id",
        additional_columns=["description"],
    )

    es.normalize_dataframe(
        new_dataframe_name="orders",
        base_dataframe_name="order_products",
        index="order_id",
        additional_columns=["customer_name", "country", "cancelled"],
    )

    es.normalize_dataframe(
        new_dataframe_name="customers",
        base_dataframe_name="orders",
        index="customer_name",
    )
    es.add_last_time_indexes()

    return es



import pandas as pd
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils.gen_utils import Library


class Median(AggregationPrimitive):
    """Determines the middlemost number in a list of values.

    Examples:
        >>> median = Median()
        >>> median([5, 3, 2, 1, 4])
        3.0

        `NaN` values are ignored.

        >>> median([5, 3, 2, 1, 4, None])
        3.0
    """

    name = "median"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    description_template = "the median of {}"

    def get_function(self, agg_type=Library.PANDAS):
        return pd.Series.median



from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from pytz import timezone

from featuretools.primitives import (
    Age,
    DateToTimeZone,
    DayOfYear,
    DaysInMonth,
    EmailAddressToDomain,
    FileExtension,
    IsFirstWeekOfMonth,
    IsFreeEmailDomain,
    IsLeapYear,
    IsLunchTime,
    IsMonthEnd,
    IsMonthStart,
    IsQuarterEnd,
    IsQuarterStart,
    IsWorkingHours,
    IsYearEnd,
    IsYearStart,
    Lag,
    NthWeekOfMonth,
    NumericLag,
    PartOfDay,
    Quarter,
    RateOfChange,
    TimeSince,
    URLToDomain,
    URLToProtocol,
    URLToTLD,
    Week,
    get_transform_primitives,
)
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


def test_time_since():
    time_since = TimeSince()
    # class datetime.datetime(year, month, day[, hour[, minute[, second[, microsecond[,
    times = pd.Series(
        [
            datetime(2019, 3, 1, 0, 0, 0, 1),
            datetime(2019, 3, 1, 0, 0, 1, 0),
            datetime(2019, 3, 1, 0, 2, 0, 0),
        ],
    )
    cutoff_time = datetime(2019, 3, 1, 0, 0, 0, 0)
    values = time_since(array=times, time=cutoff_time)

    assert list(map(int, values)) == [0, -1, -120]

    time_since = TimeSince(unit="nanoseconds")
    values = time_since(array=times, time=cutoff_time)
    assert list(map(round, values)) == [-1000, -1000000000, -120000000000]

    time_since = TimeSince(unit="milliseconds")
    values = time_since(array=times, time=cutoff_time)
    assert list(map(int, values)) == [0, -1000, -120000]

    time_since = TimeSince(unit="Milliseconds")
    values = time_since(array=times, time=cutoff_time)
    assert list(map(int, values)) == [0, -1000, -120000]

    time_since = TimeSince(unit="Years")
    values = time_since(array=times, time=cutoff_time)
    assert list(map(int, values)) == [0, 0, 0]

    times_y = pd.Series(
        [
            datetime(2019, 3, 1, 0, 0, 0, 1),
            datetime(2020, 3, 1, 0, 0, 1, 0),
            datetime(2017, 3, 1, 0, 0, 0, 0),
        ],
    )

    time_since = TimeSince(unit="Years")
    values = time_since(array=times_y, time=cutoff_time)
    assert list(map(int, values)) == [0, -1, 1]

    error_text = "Invalid unit given, make sure it is plural"
    with pytest.raises(ValueError, match=error_text):
        time_since = TimeSince(unit="na")
        time_since(array=times, time=cutoff_time)


def test_age():
    age = Age()
    dates = pd.Series(datetime(2010, 2, 26))
    ages = age(dates, time=datetime(2020, 2, 26))
    correct_ages = [10.005]  # .005 added due to leap years
    np.testing.assert_array_almost_equal(ages, correct_ages, decimal=3)


def test_age_two_years_quarterly():
    age = Age()
    dates = pd.Series(pd.date_range("2010-01-01", "2011-12-31", freq="Q"))
    ages = age(dates, time=datetime(2020, 2, 26))
    correct_ages = [9.915, 9.666, 9.414, 9.162, 8.915, 8.666, 8.414, 8.162]
    np.testing.assert_array_almost_equal(ages, correct_ages, decimal=3)


def test_age_leap_year():
    age = Age()
    dates = pd.Series([datetime(2016, 1, 1)])
    ages = age(dates, time=datetime(2016, 3, 1))
    correct_ages = [(31 + 29) / 365.0]
    np.testing.assert_array_almost_equal(ages, correct_ages, decimal=3)
    # born leap year date
    dates = pd.Series([datetime(2016, 2, 29)])
    ages = age(dates, time=datetime(2020, 2, 29))
    correct_ages = [4.0027]  # .0027 added due to leap year
    np.testing.assert_array_almost_equal(ages, correct_ages, decimal=3)


def test_age_nan():
    age = Age()
    dates = pd.Series([datetime(2010, 1, 1), np.nan, datetime(2012, 1, 1)])
    ages = age(dates, time=datetime(2020, 2, 26))
    correct_ages = [10.159, np.nan, 8.159]
    np.testing.assert_array_almost_equal(ages, correct_ages, decimal=3)


def test_day_of_year():
    doy = DayOfYear()
    dates = pd.Series([datetime(2019, 12, 31), np.nan, datetime(2020, 12, 31)])
    days_of_year = doy(dates)
    correct_days = [365, np.nan, 366]
    np.testing.assert_array_equal(days_of_year, correct_days)


def test_days_in_month():
    dim = DaysInMonth()
    dates = pd.Series(
        [datetime(2010, 1, 1), datetime(2019, 2, 1), np.nan, datetime(2020, 2, 1)],
    )
    days_in_month = dim(dates)
    correct_days = [31, 28, np.nan, 29]
    np.testing.assert_array_equal(days_in_month, correct_days)


def test_is_leap_year():
    ily = IsLeapYear()
    dates = pd.Series([datetime(2020, 1, 1), datetime(2021, 1, 1)])
    leap_year_bools = ily(dates)
    correct_bools = [True, False]
    np.testing.assert_array_equal(leap_year_bools, correct_bools)


def test_is_month_end():
    ime = IsMonthEnd()
    dates = pd.Series(
        [datetime(2019, 3, 1), datetime(2021, 2, 28), datetime(2020, 2, 29)],
    )
    ime_bools = ime(dates)
    correct_bools = [False, True, True]
    np.testing.assert_array_equal(ime_bools, correct_bools)


def test_is_month_start():
    ims = IsMonthStart()
    dates = pd.Series(
        [datetime(2019, 3, 1), datetime(2020, 2, 28), datetime(2020, 2, 29)],
    )
    ims_bools = ims(dates)
    correct_bools = [True, False, False]
    np.testing.assert_array_equal(ims_bools, correct_bools)


def test_is_quarter_end():
    iqe = IsQuarterEnd()
    dates = pd.Series([datetime(2020, 1, 1), datetime(2021, 3, 31)])
    iqe_bools = iqe(dates)
    correct_bools = [False, True]
    np.testing.assert_array_equal(iqe_bools, correct_bools)


def test_is_quarter_start():
    iqs = IsQuarterStart()
    dates = pd.Series([datetime(2020, 1, 1), datetime(2021, 3, 31)])
    iqs_bools = iqs(dates)
    correct_bools = [True, False]
    np.testing.assert_array_equal(iqs_bools, correct_bools)


def test_is_lunch_time_default():
    is_lunch_time = IsLunchTime()
    dates = pd.Series(
        [
            datetime(2022, 6, 26, 12, 12, 12),
            datetime(2022, 6, 28, 12, 3, 4),
            datetime(2022, 6, 28, 11, 3, 4),
            np.nan,
        ],
    )
    actual = is_lunch_time(dates)
    expected = [True, True, False, False]
    np.testing.assert_array_equal(actual, expected)


def test_is_lunch_time_configurable():
    is_lunch_time = IsLunchTime(14)
    dates = pd.Series(
        [
            datetime(2022, 6, 26, 12, 12, 12),
            datetime(2022, 6, 28, 14, 3, 4),
            datetime(2022, 6, 28, 11, 3, 4),
            np.nan,
        ],
    )
    actual = is_lunch_time(dates)
    expected = [False, True, False, False]
    np.testing.assert_array_equal(actual, expected)


def test_is_working_hours_standard_hours():
    is_working_hours = IsWorkingHours()
    dates = pd.Series(
        [
            datetime(2022, 6, 21, 16, 3, 3),
            datetime(2019, 1, 3, 4, 4, 4),
            datetime(2022, 1, 1, 12, 1, 2),
        ],
    )
    actual = is_working_hours(dates).tolist()
    expected = [True, False, True]
    np.testing.assert_array_equal(actual, expected)


def test_is_working_hours_configured_hours():
    is_working_hours = IsWorkingHours(15, 18)
    dates = pd.Series(
        [
            datetime(2022, 6, 21, 16, 3, 3),
            datetime(2022, 6, 26, 14, 4, 4),
            datetime(2022, 1, 1, 12, 1, 2),
        ],
    )
    answer = is_working_hours(dates).tolist()
    expected = [True, False, False]
    np.testing.assert_array_equal(answer, expected)


def test_part_of_day():
    pod = PartOfDay()
    dates = pd.Series(
        [
            datetime(2020, 1, 11, 0, 2, 1),
            datetime(2020, 1, 11, 1, 2, 1),
            datetime(2021, 3, 31, 4, 2, 1),
            datetime(2020, 3, 4, 6, 2, 1),
            datetime(2020, 3, 4, 8, 2, 1),
            datetime(2020, 3, 4, 11, 2, 1),
            datetime(2020, 3, 4, 14, 2, 3),
            datetime(2020, 3, 4, 17, 2, 3),
            datetime(2020, 2, 2, 20, 2, 2),
            np.nan,
        ],
    )
    actual = pod(dates)
    expected = pd.Series(
        [
            "midnight",
            "midnight",
            "dawn",
            "early morning",
            "late morning",
            "noon",
            "afternoon",
            "evening",
            "night",
            np.nan,
        ],
    )
    pd.testing.assert_series_equal(expected, actual)


def test_is_year_end():
    is_year_end = IsYearEnd()
    dates = pd.Series([datetime(2020, 12, 31), np.nan, datetime(2020, 1, 1)])
    answer = is_year_end(dates)
    correct_answer = [True, False, False]
    np.testing.assert_array_equal(answer, correct_answer)


def test_is_year_start():
    is_year_start = IsYearStart()
    dates = pd.Series([datetime(2020, 12, 31), np.nan, datetime(2020, 1, 1)])
    answer = is_year_start(dates)
    correct_answer = [False, False, True]
    np.testing.assert_array_equal(answer, correct_answer)


def test_quarter_regular():
    q = Quarter()
    array = pd.Series(
        [
            pd.to_datetime("2018-01-01"),
            pd.to_datetime("2018-04-01"),
            pd.to_datetime("2018-07-01"),
            pd.to_datetime("2018-10-01"),
        ],
    )
    answer = q(array)
    correct_answer = pd.Series([1, 2, 3, 4])
    np.testing.assert_array_equal(answer, correct_answer)


def test_quarter_leap_year():
    q = Quarter()
    array = pd.Series(
        [
            pd.to_datetime("2016-02-29"),
            pd.to_datetime("2018-04-01"),
            pd.to_datetime("2018-07-01"),
            pd.to_datetime("2018-10-01"),
        ],
    )
    answer = q(array)
    correct_answer = pd.Series([1, 2, 3, 4])
    np.testing.assert_array_equal(answer, correct_answer)


def test_quarter_nan_and_nat_input():
    q = Quarter()
    array = pd.Series(
        [
            pd.to_datetime("2016-02-29"),
            np.nan,
            np.datetime64("NaT"),
            pd.to_datetime("2018-10-01"),
        ],
    )
    answer = q(array)
    correct_answer = pd.Series([1, np.nan, np.nan, 4])
    np.testing.assert_array_equal(answer, correct_answer)


def test_quarter_year_before_1970():
    q = Quarter()
    array = pd.Series(
        [
            pd.to_datetime("2018-01-01"),
            pd.to_datetime("1950-04-01"),
            pd.to_datetime("1874-07-01"),
            pd.to_datetime("2018-10-01"),
        ],
    )
    answer = q(array)
    correct_answer = pd.Series([1, 2, 3, 4])
    np.testing.assert_array_equal(answer, correct_answer)


def test_quarter_year_after_2038():
    q = Quarter()
    array = pd.Series(
        [
            pd.to_datetime("2018-01-01"),
            pd.to_datetime("2050-04-01"),
            pd.to_datetime("2174-07-01"),
            pd.to_datetime("2018-10-01"),
        ],
    )
    answer = q(array)
    correct_answer = pd.Series([1, 2, 3, 4])
    np.testing.assert_array_equal(answer, correct_answer)


def test_quarter():
    q = Quarter()
    dates = [datetime(2019, 12, 1), datetime(2019, 1, 3), datetime(2020, 2, 1)]
    quarter = q(dates)
    correct_quarters = [4, 1, 1]
    np.testing.assert_array_equal(quarter, correct_quarters)


def test_week_no_deprecation_message():
    dates = [
        datetime(2019, 1, 3),
        datetime(2019, 6, 17, 11, 10, 50),
        datetime(2019, 11, 30, 19, 45, 15),
    ]
    with pytest.warns(None) as record:
        week = Week()
        week(dates).tolist()
    assert not record


def test_url_to_domain_urls():
    url_to_domain = URLToDomain()
    urls = pd.Series(
        [
            "https://play.google.com/store/apps/details?id=com.skgames.trafficracer%22",
            "http://mplay.google.co.in/sadfask/asdkfals?dk=10",
            "http://lplay.google.co.in/sadfask/asdkfals?dk=10",
            "http://play.google.co.in/sadfask/asdkfals?dk=10",
            "http://tplay.google.co.in/sadfask/asdkfals?dk=10",
            "http://www.google.co.in/sadfask/asdkfals?dk=10",
            "www.google.co.in/sadfask/asdkfals?dk=10",
            "http://user:pass@google.com/?a=b#asdd",
            "https://www.compzets.com?asd=10",
            "www.compzets.com?asd=10",
            "facebook.com",
            "https://www.compzets.net?asd=10",
            "http://www.featuretools.org",
        ],
    )
    correct_urls = [
        "play.google.com",
        "mplay.google.co.in",
        "lplay.google.co.in",
        "play.google.co.in",
        "tplay.google.co.in",
        "google.co.in",
        "google.co.in",
        "google.com",
        "compzets.com",
        "compzets.com",
        "facebook.com",
        "compzets.net",
        "featuretools.org",
    ]
    np.testing.assert_array_equal(url_to_domain(urls), correct_urls)


def test_url_to_domain_long_url():
    url_to_domain = URLToDomain()
    urls = pd.Series(
        [
            "http://chart.apis.google.com/chart?chs=500x500&chma=0,0,100, \
                        100&cht=p&chco=FF0000%2CFFFF00%7CFF8000%2C00FF00%7C00FF00%2C0 \
                        000FF&chd=t%3A122%2C42%2C17%2C10%2C8%2C7%2C7%2C7%2C7%2C6%2C6% \
                        2C6%2C6%2C5%2C5&chl=122%7C42%7C17%7C10%7C8%7C7%7C7%7C7%7C7%7C \
                        6%7C6%7C6%7C6%7C5%7C5&chdl=android%7Cjava%7Cstack-trace%7Cbro \
                        adcastreceiver%7Candroid-ndk%7Cuser-agent%7Candroid-webview%7 \
                        Cwebview%7Cbackground%7Cmultithreading%7Candroid-source%7Csms \
                        %7Cadb%7Csollections%7Cactivity|Chart",
        ],
    )
    correct_urls = ["chart.apis.google.com"]
    results = url_to_domain(urls)
    np.testing.assert_array_equal(results, correct_urls)


def test_url_to_domain_nan():
    url_to_domain = URLToDomain()
    urls = pd.Series(["www.featuretools.com", np.nan], dtype="object")
    correct_urls = pd.Series(["featuretools.com", np.nan], dtype="object")
    results = url_to_domain(urls)
    pd.testing.assert_series_equal(results, correct_urls)


def test_url_to_protocol_urls():
    url_to_protocol = URLToProtocol()
    urls = pd.Series(
        [
            "https://play.google.com/store/apps/details?id=com.skgames.trafficracer%22",
            "http://mplay.google.co.in/sadfask/asdkfals?dk=10",
            "http://lplay.google.co.in/sadfask/asdkfals?dk=10",
            "www.google.co.in/sadfask/asdkfals?dk=10",
            "http://user:pass@google.com/?a=b#asdd",
            "https://www.compzets.com?asd=10",
            "www.compzets.com?asd=10",
            "facebook.com",
            "https://www.compzets.net?asd=10",
            "http://www.featuretools.org",
            "https://featuretools.com",
        ],
    )
    correct_urls = pd.Series(
        [
            "https",
            "http",
            "http",
            np.nan,
            "http",
            "https",
            np.nan,
            np.nan,
            "https",
            "http",
            "https",
        ],
    )
    results = url_to_protocol(urls)
    pd.testing.assert_series_equal(results, correct_urls)


def test_url_to_protocol_long_url():
    url_to_protocol = URLToProtocol()
    urls = pd.Series(
        [
            "http://chart.apis.google.com/chart?chs=500x500&chma=0,0,100, \
                        100&cht=p&chco=FF0000%2CFFFF00%7CFF8000%2C00FF00%7C00FF00%2C0 \
                        000FF&chd=t%3A122%2C42%2C17%2C10%2C8%2C7%2C7%2C7%2C7%2C6%2C6% \
                        2C6%2C6%2C5%2C5&chl=122%7C42%7C17%7C10%7C8%7C7%7C7%7C7%7C7%7C \
                        6%7C6%7C6%7C6%7C5%7C5&chdl=android%7Cjava%7Cstack-trace%7Cbro \
                        adcastreceiver%7Candroid-ndk%7Cuser-agent%7Candroid-webview%7 \
                        Cwebview%7Cbackground%7Cmultithreading%7Candroid-source%7Csms \
                        %7Cadb%7Csollections%7Cactivity|Chart",
        ],
    )
    correct_urls = ["http"]
    results = url_to_protocol(urls)
    np.testing.assert_array_equal(results, correct_urls)


def test_url_to_protocol_nan():
    url_to_protocol = URLToProtocol()
    urls = pd.Series(["www.featuretools.com", np.nan, ""], dtype="object")
    correct_urls = pd.Series([np.nan, np.nan, np.nan], dtype="object")
    results = url_to_protocol(urls)
    pd.testing.assert_series_equal(results, correct_urls)


def test_url_to_tld_urls():
    url_to_tld = URLToTLD()
    urls = pd.Series(
        [
            "https://play.google.com/store/apps/details?id=com.skgames.trafficracer%22",
            "http://mplay.google.co.in/sadfask/asdkfals?dk=10",
            "http://lplay.google.co.in/sadfask/asdkfals?dk=10",
            "http://play.google.co.in/sadfask/asdkfals?dk=10",
            "http://tplay.google.co.in/sadfask/asdkfals?dk=10",
            "http://www.google.co.in/sadfask/asdkfals?dk=10",
            "www.google.co.in/sadfask/asdkfals?dk=10",
            "http://user:pass@google.com/?a=b#asdd",
            "https://www.compzets.dev?asd=10",
            "www.compzets.com?asd=10",
            "https://www.compzets.net?asd=10",
            "http://www.featuretools.org",
            "featuretools.org",
        ],
    )
    correct_urls = [
        "com",
        "in",
        "in",
        "in",
        "in",
        "in",
        "in",
        "com",
        "dev",
        "com",
        "net",
        "org",
        "org",
    ]
    np.testing.assert_array_equal(url_to_tld(urls), correct_urls)


def test_url_to_tld_long_url():
    url_to_tld = URLToTLD()
    urls = pd.Series(
        [
            "http://chart.apis.google.com/chart?chs=500x500&chma=0,0,100, \
                        100&cht=p&chco=FF0000%2CFFFF00%7CFF8000%2C00FF00%7C00FF00%2C0 \
                        000FF&chd=t%3A122%2C42%2C17%2C10%2C8%2C7%2C7%2C7%2C7%2C6%2C6% \
                        2C6%2C6%2C5%2C5&chl=122%7C42%7C17%7C10%7C8%7C7%7C7%7C7%7C7%7C \
                        6%7C6%7C6%7C6%7C5%7C5&chdl=android%7Cjava%7Cstack-trace%7Cbro \
                        adcastreceiver%7Candroid-ndk%7Cuser-agent%7Candroid-webview%7 \
                        Cwebview%7Cbackground%7Cmultithreading%7Candroid-source%7Csms \
                        %7Cadb%7Csollections%7Cactivity|Chart",
        ],
    )
    correct_urls = ["com"]
    np.testing.assert_array_equal(url_to_tld(urls), correct_urls)


def test_url_to_tld_nan():
    url_to_tld = URLToTLD()
    urls = pd.Series(
        ["www.featuretools.com", np.nan, "featuretools", ""],
        dtype="object",
    )
    correct_urls = pd.Series(["com", np.nan, np.nan, np.nan], dtype="object")
    results = url_to_tld(urls)
    pd.testing.assert_series_equal(results, correct_urls, check_names=False)


def test_is_free_email_domain_valid_addresses():
    is_free_email_domain = IsFreeEmailDomain()
    array = pd.Series(
        [
            "test@hotmail.com",
            "name@featuretools.com",
            "nobody@yahoo.com",
            "free@gmail.com",
        ],
    )
    answers = pd.Series(is_free_email_domain(array))
    correct_answers = pd.Series([True, False, True, True])
    pd.testing.assert_series_equal(answers, correct_answers)


def test_is_free_email_domain_valid_addresses_whitespace():
    is_free_email_domain = IsFreeEmailDomain()
    array = pd.Series(
        [
            " test@hotmail.com",
            " name@featuretools.com",
            "nobody@yahoo.com ",
            " free@gmail.com ",
        ],
    )
    answers = pd.Series(is_free_email_domain(array))
    correct_answers = pd.Series([True, False, True, True])
    pd.testing.assert_series_equal(answers, correct_answers)


def test_is_free_email_domain_nan():
    is_free_email_domain = IsFreeEmailDomain()
    array = pd.Series([np.nan, "name@featuretools.com", "nobody@yahoo.com"])
    answers = pd.Series(is_free_email_domain(array))
    correct_answers = pd.Series([np.nan, False, True])
    pd.testing.assert_series_equal(answers, correct_answers)


def test_is_free_email_domain_empty_string():
    is_free_email_domain = IsFreeEmailDomain()
    array = pd.Series(["", "name@featuretools.com", "nobody@yahoo.com"])
    answers = pd.Series(is_free_email_domain(array))
    correct_answers = pd.Series([np.nan, False, True])
    pd.testing.assert_series_equal(answers, correct_answers)


def test_is_free_email_domain_empty_series():
    is_free_email_domain = IsFreeEmailDomain()
    array = pd.Series([], dtype="category")
    answers = pd.Series(is_free_email_domain(array))
    correct_answers = pd.Series([], dtype="category")
    pd.testing.assert_series_equal(answers, correct_answers)


def test_is_free_email_domain_invalid_email():
    is_free_email_domain = IsFreeEmailDomain()
    array = pd.Series(
        [
            np.nan,
            "this is not an email address",
            "name@featuretools.com",
            "nobody@yahoo.com",
            1234,
            1.23,
            True,
        ],
    )
    answers = pd.Series(is_free_email_domain(array))
    correct_answers = pd.Series([np.nan, np.nan, False, True, np.nan, np.nan, np.nan])
    pd.testing.assert_series_equal(answers, correct_answers)


def test_is_free_email_domain_all_nan():
    is_free_email_domain = IsFreeEmailDomain()
    array = pd.Series([np.nan, np.nan])
    answers = pd.Series(is_free_email_domain(array))
    correct_answers = pd.Series([np.nan, np.nan], dtype=object)
    pd.testing.assert_series_equal(answers, correct_answers)


def test_email_address_to_domain_valid_addresses():
    email_address_to_domain = EmailAddressToDomain()
    array = pd.Series(
        [
            "test@hotmail.com",
            "name@featuretools.com",
            "nobody@yahoo.com",
            "free@gmail.com",
        ],
    )
    answers = pd.Series(email_address_to_domain(array))
    correct_answers = pd.Series(
        ["hotmail.com", "featuretools.com", "yahoo.com", "gmail.com"],
    )
    pd.testing.assert_series_equal(answers, correct_answers)


def test_email_address_to_domain_valid_addresses_whitespace():
    email_address_to_domain = EmailAddressToDomain()
    array = pd.Series(
        [
            " test@hotmail.com",
            " name@featuretools.com",
            "nobody@yahoo.com ",
            " free@gmail.com ",
        ],
    )
    answers = pd.Series(email_address_to_domain(array))
    correct_answers = pd.Series(
        ["hotmail.com", "featuretools.com", "yahoo.com", "gmail.com"],
    )
    pd.testing.assert_series_equal(answers, correct_answers)


def test_email_address_to_domain_nan():
    email_address_to_domain = EmailAddressToDomain()
    array = pd.Series([np.nan, "name@featuretools.com", "nobody@yahoo.com"])
    answers = pd.Series(email_address_to_domain(array))
    correct_answers = pd.Series([np.nan, "featuretools.com", "yahoo.com"])
    pd.testing.assert_series_equal(answers, correct_answers)


def test_email_address_to_domain_empty_string():
    email_address_to_domain = EmailAddressToDomain()
    array = pd.Series(["", "name@featuretools.com", "nobody@yahoo.com"])
    answers = pd.Series(email_address_to_domain(array))
    correct_answers = pd.Series([np.nan, "featuretools.com", "yahoo.com"])
    pd.testing.assert_series_equal(answers, correct_answers)


def test_email_address_to_domain_empty_series():
    email_address_to_domain = EmailAddressToDomain()
    array = pd.Series([], dtype="category")
    answers = pd.Series(email_address_to_domain(array))
    correct_answers = pd.Series([], dtype="category")
    pd.testing.assert_series_equal(answers, correct_answers)


def test_email_address_to_domain_invalid_email():
    email_address_to_domain = EmailAddressToDomain()
    array = pd.Series(
        [
            np.nan,
            "this is not an email address",
            "name@featuretools.com",
            "nobody@yahoo.com",
            1234,
            1.23,
            True,
        ],
    )
    answers = pd.Series(email_address_to_domain(array))
    correct_answers = pd.Series(
        [np.nan, np.nan, "featuretools.com", "yahoo.com", np.nan, np.nan, np.nan],
    )
    pd.testing.assert_series_equal(answers, correct_answers)


def test_email_address_to_domain_all_nan():
    email_address_to_domain = EmailAddressToDomain()
    array = pd.Series([np.nan, np.nan])
    answers = pd.Series(email_address_to_domain(array))
    correct_answers = pd.Series([np.nan, np.nan], dtype=object)
    pd.testing.assert_series_equal(answers, correct_answers)


def test_trans_primitives_can_init_without_params():
    trans_primitives = get_transform_primitives().values()
    for trans_primitive in trans_primitives:
        trans_primitive()


def test_numeric_lag_future_warning():
    warning_text = "NumericLag is deprecated and will be removed in a future version. Please use the 'Lag' primitive instead."
    with pytest.warns(FutureWarning, match=warning_text):
        NumericLag()


def test_lag_regular():
    primitive_instance = Lag()
    primitive_func = primitive_instance.get_function()

    array = pd.Series([1, 2, 3, 4])
    time_array = pd.Series(pd.date_range(start="2020-01-01", periods=4, freq="D"))

    answer = pd.Series(primitive_func(array, time_array))

    correct_answer = pd.Series([np.nan, 1, 2, 3])
    pd.testing.assert_series_equal(answer, correct_answer)


def test_lag_period():
    primitive_instance = Lag(periods=3)
    primitive_func = primitive_instance.get_function()

    array = pd.Series([1, 2, 3, 4])
    time_array = pd.Series(pd.date_range(start="2020-01-01", periods=4, freq="D"))

    answer = pd.Series(primitive_func(array, time_array))

    correct_answer = pd.Series([np.nan, np.nan, np.nan, 1])
    pd.testing.assert_series_equal(answer, correct_answer)


def test_lag_negative_period():
    primitive_instance = Lag(periods=-2)
    primitive_func = primitive_instance.get_function()

    array = pd.Series([1, 2, 3, 4])
    time_array = pd.Series(pd.date_range(start="2020-01-01", periods=4, freq="D"))

    answer = pd.Series(primitive_func(array, time_array))

    correct_answer = pd.Series([3, 4, np.nan, np.nan])
    pd.testing.assert_series_equal(answer, correct_answer)


def test_lag_starts_with_nan():
    primitive_instance = Lag()
    primitive_func = primitive_instance.get_function()

    array = pd.Series([np.nan, 2, 3, 4])
    time_array = pd.Series(pd.date_range(start="2020-01-01", periods=4, freq="D"))

    answer = pd.Series(primitive_func(array, time_array))

    correct_answer = pd.Series([np.nan, np.nan, 2, 3])
    pd.testing.assert_series_equal(answer, correct_answer)


def test_lag_ends_with_nan():
    primitive_instance = Lag()
    primitive_func = primitive_instance.get_function()

    array = pd.Series([1, 2, 3, np.nan])
    time_array = pd.Series(pd.date_range(start="2020-01-01", periods=4, freq="D"))

    answer = pd.Series(primitive_func(array, time_array))

    correct_answer = pd.Series([np.nan, 1, 2, 3])
    pd.testing.assert_series_equal(answer, correct_answer)


@pytest.mark.parametrize(
    "input_array,expected_output",
    [
        (
            pd.Series(["hello", "world", "foo", "bar"], dtype="string"),
            pd.Series([np.nan, "hello", "world", "foo"], dtype="string"),
        ),
        (
            pd.Series(["cow", "cow", "pig", "pig"], dtype="category"),
            pd.Series([np.nan, "cow", "cow", "pig"], dtype="category"),
        ),
        (
            pd.Series([True, False, True, False], dtype="bool"),
            pd.Series([np.nan, True, False, True], dtype="object"),
        ),
        (
            pd.Series([True, False, True, False], dtype="boolean"),
            pd.Series([np.nan, True, False, True], dtype="boolean"),
        ),
        (
            pd.Series([1.23, 2.45, 3.56, 4.98], dtype="float"),
            pd.Series([np.nan, 1.23, 2.45, 3.56], dtype="float"),
        ),
        (
            pd.Series([1, 2, 3, 4], dtype="Int64"),
            pd.Series([np.nan, 1, 2, 3], dtype="Int64"),
        ),
        (
            pd.Series([1, 2, 3, 4], dtype="int64"),
            pd.Series([np.nan, 1, 2, 3], dtype="float64"),
        ),
    ],
)
def test_lag_with_different_dtypes(input_array, expected_output):
    primitive_instance = Lag()
    primitive_func = primitive_instance.get_function()
    time_array = pd.Series(pd.date_range(start="2020-01-01", periods=4, freq="D"))
    answer = pd.Series(primitive_func(input_array, time_array))
    pd.testing.assert_series_equal(answer, expected_output)


def test_date_to_time_zone_primitive():
    primitive_func = DateToTimeZone().get_function()
    x = pd.Series(
        [
            datetime(2010, 1, 1, tzinfo=timezone("America/Los_Angeles")),
            datetime(2010, 1, 10, tzinfo=timezone("Singapore")),
            datetime(2020, 1, 1, tzinfo=timezone("UTC")),
            datetime(2010, 1, 1, tzinfo=timezone("Europe/London")),
        ],
    )
    answer = pd.Series(["America/Los_Angeles", "Singapore", "UTC", "Europe/London"])
    pd.testing.assert_series_equal(primitive_func(x), answer)


def test_date_to_time_zone_datetime64():
    primitive_func = DateToTimeZone().get_function()
    x = pd.Series(
        [
            datetime(2010, 1, 1),
            datetime(2010, 1, 10),
            datetime(2020, 1, 1),
        ],
    ).astype("datetime64[ns]")
    x = x.dt.tz_localize("America/Los_Angeles")
    answer = pd.Series(["America/Los_Angeles"] * 3)
    pd.testing.assert_series_equal(primitive_func(x), answer)


def test_date_to_time_zone_naive_dates():
    primitive_func = DateToTimeZone().get_function()
    x = pd.Series(
        [
            datetime(2010, 1, 1, tzinfo=timezone("America/Los_Angeles")),
            datetime(2010, 1, 1),
            datetime(2010, 1, 2),
        ],
    )
    answer = pd.Series(["America/Los_Angeles", np.nan, np.nan])
    pd.testing.assert_series_equal(primitive_func(x), answer)


def test_date_to_time_zone_nan():
    primitive_func = DateToTimeZone().get_function()
    x = pd.Series(
        [
            datetime(2010, 1, 1, tzinfo=timezone("America/Los_Angeles")),
            pd.NaT,
            np.nan,
        ],
    )
    answer = pd.Series(["America/Los_Angeles", np.nan, np.nan])
    pd.testing.assert_series_equal(primitive_func(x), answer)


def test_rate_of_change_primitive_regular_interval():
    rate_of_change = RateOfChange()
    times = pd.date_range(start="2019-01-01", freq="2s", periods=5)
    values = [0, 30, 180, -90, 0]
    expected = pd.Series([np.nan, 15, 75, -135, 45])
    actual = rate_of_change(values, times)
    pd.testing.assert_series_equal(actual, expected)


def test_rate_of_change_primitive_uneven_interval():
    rate_of_change = RateOfChange()
    times = pd.to_datetime(
        [
            "2019-01-01 00:00:00",
            "2019-01-01 00:00:01",
            "2019-01-01 00:00:03",
            "2019-01-01 00:00:07",
            "2019-01-01 00:00:08",
        ],
    )
    values = [0, 30, 180, -90, 0]
    expected = pd.Series([np.nan, 30, 75, -67.5, 90])
    actual = rate_of_change(values, times)
    pd.testing.assert_series_equal(actual, expected)


def test_rate_of_change_primitive_with_nan():
    rate_of_change = RateOfChange()
    times = pd.date_range(start="2019-01-01", freq="2s", periods=5)
    values = [0, 30, np.nan, -90, 0]
    expected = pd.Series([np.nan, 15, np.nan, np.nan, 45])
    actual = rate_of_change(values, times)
    pd.testing.assert_series_equal(actual, expected)


class TestFileExtension(PrimitiveTestBase):
    primitive = FileExtension

    def test_filepaths(self):
        primitive_func = FileExtension().get_function()
        array = pd.Series(
            [
                "doc.txt",
                "~/documents/data.json",
                "data.JSON",
                "C:\\Projects\\apilibrary\\apilibrary.sln",
            ],
            dtype="string",
        )
        answer = pd.Series([".txt", ".json", ".json", ".sln"], dtype="string")
        pd.testing.assert_series_equal(primitive_func(array), answer)

    def test_invalid(self):
        primitive_func = FileExtension().get_function()
        array = pd.Series(["doc.txt", "~/documents/data", np.nan], dtype="string")
        answer = pd.Series([".txt", np.nan, np.nan], dtype="string")
        pd.testing.assert_series_equal(primitive_func(array), answer)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(
            es,
            aggregation,
            transform,
            self.primitive,
            target_dataframe_name="sessions",
        )


class TestIsFirstWeekOfMonth(PrimitiveTestBase):
    primitive = IsFirstWeekOfMonth

    def test_valid_dates(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("03/01/2019"),
                pd.to_datetime("03/03/2019"),
                pd.to_datetime("03/31/2019"),
                pd.to_datetime("03/30/2019"),
            ],
        )
        answers = primitive_func(array).tolist()
        correct_answers = [True, False, False, False]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_leap_year(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("03/01/2019"),
                pd.to_datetime("02/29/2016"),
                pd.to_datetime("03/31/2019"),
                pd.to_datetime("03/30/2019"),
            ],
        )
        answers = primitive_func(array).tolist()
        correct_answers = [True, False, False, False]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_year_before_1970(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("06/01/1965"),
                pd.to_datetime("03/02/2019"),
                pd.to_datetime("03/31/2019"),
                pd.to_datetime("03/30/2019"),
            ],
        )
        answers = primitive_func(array).tolist()
        correct_answers = [True, True, False, False]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_year_after_2038(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("12/31/2040"),
                pd.to_datetime("01/01/2040"),
                pd.to_datetime("03/31/2019"),
                pd.to_datetime("03/30/2019"),
            ],
        )
        answers = primitive_func(array).tolist()
        correct_answers = [False, True, False, False]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_nan_input(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("03/01/2019"),
                np.nan,
                np.datetime64("NaT"),
                pd.to_datetime("03/30/2019"),
            ],
        )
        answers = primitive_func(array).tolist()
        correct_answers = [True, np.nan, np.nan, False]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)


class TestNthWeekOfMonth(PrimitiveTestBase):
    primitive = NthWeekOfMonth

    def test_valid_dates(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("03/01/2019"),
                pd.to_datetime("03/03/2019"),
                pd.to_datetime("03/31/2019"),
                pd.to_datetime("03/30/2019"),
                pd.to_datetime("09/01/2019"),
            ],
        )
        answers = primitive_func(array)
        correct_answers = [1, 2, 6, 5, 1]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_leap_year(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("03/01/2019"),
                pd.to_datetime("02/29/2016"),
                pd.to_datetime("03/31/2019"),
                pd.to_datetime("03/30/2019"),
            ],
        )
        answers = primitive_func(array)
        correct_answers = [1, 5, 6, 5]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_year_before_1970(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("06/06/1965"),
                pd.to_datetime("03/02/2019"),
                pd.to_datetime("03/31/2019"),
                pd.to_datetime("03/30/2019"),
            ],
        )
        answers = primitive_func(array)
        correct_answers = [2, 1, 6, 5]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_year_after_2038(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("12/31/2040"),
                pd.to_datetime("01/01/2001"),
                pd.to_datetime("03/31/2019"),
                pd.to_datetime("03/30/2019"),
            ],
        )
        answers = primitive_func(array)
        correct_answers = [6, 1, 6, 5]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_nan_input(self):
        primitive_func = self.primitive().get_function()
        array = pd.Series(
            [
                pd.to_datetime("03/01/2019"),
                np.nan,
                np.datetime64("NaT"),
                pd.to_datetime("03/30/2019"),
            ],
        )
        answers = primitive_func(array)
        correct_answers = [1, np.nan, np.nan, 5]
        np.testing.assert_array_equal(answers, correct_answers)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



import inspect
from collections import defaultdict
from itertools import combinations, permutations, product
from typing import Iterable, List, Set, Tuple, Type, Union, cast

from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import LogicalType
from woodwork.table_schema import TableSchema

from featuretools.feature_discovery.FeatureCollection import FeatureCollection
from featuretools.feature_discovery.LiteFeature import LiteFeature
from featuretools.feature_discovery.utils import column_schema_to_keys
from featuretools.primitives.base.primitive_base import PrimitiveBase
from featuretools.tests.testing_utils.generate_fake_dataframe import flatten_list


def _index_column_set(column_set: List[ColumnSchema]) -> List[Tuple[str, int]]:
    """
    Indexes input set to find types of columns and the quantity of each

    Args:
        column_set (List(ColumnSchema)):
            List of Column types needed by associated primitive.

    Returns:
        List[Tuple[str, int]]
            A list of key, count tuples

    Examples:
        .. code-block:: python

            from featuretools.feature_discovery.feature_discovery import _index_column_set
            from woodwork.column_schema import ColumnSchema

            column_set = [ColumnSchema(semantic_tags={"numeric"}), ColumnSchema(semantic_tags={"numeric"})]
            indexed_column_set = _index_column_set(column_set)
            [("numeric": 2)]
    """
    out = defaultdict(int)
    for column_schema in column_set:
        key = column_schema_to_keys(column_schema)
        out[key] += 1
    return list(out.items())


def _get_features(
    feature_collection: FeatureCollection,
    column_keys: Tuple[Tuple[str, int]],
    commutative: bool,
) -> List[List[LiteFeature]]:
    """
    Calculates all LiteFeature combinations using the given hashmap of existing features, and the input set of required columns.

    Args:
        feature_collection (FeatureCollection):
            An indexed feature collection object for efficient querying of features
        column_keys (List[Tuple[str, int]]):
            List of Column types needed by associated primitive.
        commutative (bool):
            whether or not we need to use product or combinations to create feature sets.

    Returns:
        List[List[LiteFeature]]
            A list of LiteFeature sets.

    Examples:
        .. code-block:: python

            from featuretools.feature_discovery.feature_discovery import _get_features
            from woodwork.column_schema import ColumnSchema

            feature_groups = {
                "ANY": ["f1", "f2", "f3"],
                "Double": ["f1", "f2", "f3"],
                "numeric": ["f1", "f2", "f3"],
                "Double,numeric": ["f1", "f2", "f3"],
            }
            column_set = [ColumnSchema(semantic_tags={"numeric"}), ColumnSchema(semantic_tags={"numeric"})]
            features = _get_features(col_groups, column_set, commutative=False)
    """

    prod_iter = []
    for key, count in column_keys:
        relevant_features = list(feature_collection.get_by_key(key))

        if commutative:
            prod_iter.append(combinations(relevant_features, count))
        else:
            prod_iter.append(permutations(relevant_features, count))

    feature_combinations = product(*prod_iter)

    return [flatten_list(x) for x in feature_combinations]


def _primitive_to_columnsets(primitive: PrimitiveBase) -> List[List[ColumnSchema]]:
    column_sets = primitive.input_types
    assert column_sets is not None
    if not isinstance(column_sets[0], list):
        column_sets = [primitive.input_types]

    column_sets = cast(List[List[ColumnSchema]], column_sets)

    # Some primitives are commutative, yet have explicit versions of commutative pairs (eg. MultiplyNumericBoolean),
    # which would create multiple versions, so this resolved that.
    if primitive.commutative:
        existing = set()
        uniq_column_sets = []
        for column_set in column_sets:
            key = "_".join(sorted([x.__repr__() for x in column_set]))
            if key not in existing:
                uniq_column_sets.append(column_set)
                existing.add(key)

        column_sets = uniq_column_sets

    return column_sets


def _get_matching_features(
    feature_collection: FeatureCollection,
    primitive: PrimitiveBase,
) -> List[List[LiteFeature]]:
    """
    For a given primitive, find all feature sets that can be used to create new feature

    Args:
        feature_collection (FeatureCollection):
            An indexed feature collection object for efficient querying of features
        primitive (PrimitiveBase)

    Returns:
        List[List[LiteFeature]]
            List of feature sets

    Examples:
        .. code-block:: python

            from featuretools.feature_discovery.feature_discovery import get_matching_columns
            from woodwork.column_schema import ColumnSchema

            feature_groups = {
                "ANY": ["f1", "f2", "f3"],
                "Double": ["f1", "f2", "f3"],
                "numeric": ["f1", "f2", "f3"],
                "Double,numeric": ["f1", "f2", "f3"],
            }

            feature_sets = _get_matching_features(col_groups, AddNumeric)

            [
                ["f1", "f2"],
                ["f1", "f3"],
                ["f2", "f3"]
            ]
    """
    column_sets = _primitive_to_columnsets(primitive=primitive)

    column_keys_set = [_index_column_set(c) for c in column_sets]

    commutative = primitive.commutative

    feature_sets = []
    for column_keys in column_keys_set:
        assert column_keys is not None
        feature_sets_ = _get_features(
            feature_collection=feature_collection,
            column_keys=tuple(column_keys),
            commutative=commutative,
        )

        feature_sets.extend(feature_sets_)

    return feature_sets


def _features_from_primitive(
    primitive: PrimitiveBase,
    feature_collection: FeatureCollection,
) -> List[LiteFeature]:
    """
    For a given primitive, creates all engineered features

    Args:
        primitive (Type[PrimitiveBase])
        feature_collection (FeatureCollection):
            An indexed feature collection object for efficient querying of features

    Returns:
        List[List[LiteFeature]]
            List of feature sets

    Examples:
        .. code-block:: python

            from featuretools.feature_discovery.feature_discovery import get_matching_columns
            from woodwork.column_schema import ColumnSchema

            feature_groups = {
                "ANY": ["f1", "f2", "f3"],
                "Double": ["f1", "f2", "f3"],
                "numeric": ["f1", "f2", "f3"],
                "Double,numeric": ["f1", "f2", "f3"],
            }

            feature_sets = _features_from_primitive(AddNumeric, feature_groups)

            [
                ["f1", "f2"],
                ["f1", "f3"],
                ["f2", "f3"]
            ]
    """
    assert isinstance(primitive, PrimitiveBase)

    features: List[LiteFeature] = []
    feature_sets = _get_matching_features(
        feature_collection=feature_collection,
        primitive=primitive,
    )
    for feature_set in feature_sets:
        if primitive.number_output_features > 1:
            related_features: Set[LiteFeature] = set()
            for n in range(primitive.number_output_features):
                feature = LiteFeature(
                    primitive=primitive,
                    base_features=feature_set,
                    idx=n,
                )

                related_features.add(feature)

            for f in related_features:
                f.related_features = related_features - {f}
                features.append(f)
        else:
            features.append(
                LiteFeature(
                    primitive=primitive,
                    base_features=feature_set,
                ),
            )
    return features


def schema_to_features(schema: TableSchema) -> List[LiteFeature]:
    """
    ** EXPERIMENTAL **
    Convert a Woodwork Schema object to a list of LiteFeatures.

    Args:
        schema (TableSchema):
            Woodwork TableSchema object

    Returns:
        List[LiteFeature]

    Examples:
        .. code-block:: python

            from featuretools.feature_discovery.feature_discovery import schema_to_features
            from featuretools.primitives import Absolute, IsNull
            import pandas as pd
            import woodwork as ww

            df = pd.DataFrame({
                "idx": [0,1,2,3],
                "f1": ["A", "B", "C", "D"],
                "f2": [1.2, 2.3, 3.4, 4.5]
            })

            df.ww.init()

            features = schema_to_features(df.ww.schema)

    """
    features = []
    for col_name, column_schema in schema.columns.items():
        assert isinstance(column_schema, ColumnSchema)

        logical_type = column_schema.logical_type
        assert logical_type
        assert issubclass(type(logical_type), LogicalType)

        tags = column_schema.semantic_tags
        assert isinstance(tags, set)

        features.append(
            LiteFeature(
                name=col_name,
                logical_type=type(logical_type),
                tags=tags,
            ),
        )

    return features


def _check_inputs(
    input_features: Iterable[LiteFeature],
    primitives: Union[List[Type[PrimitiveBase]], List[PrimitiveBase]],
) -> Tuple[Iterable[LiteFeature], List[PrimitiveBase]]:
    if not isinstance(input_features, Iterable):
        raise ValueError("input_features must be an iterable of LiteFeature objects")

    for feature in input_features:
        if not isinstance(feature, LiteFeature):
            raise ValueError(
                "input_features must be an iterable of LiteFeature objects",
            )

    if not isinstance(primitives, List):
        raise ValueError(
            "primitives must be a list of Primitive classes or Primitive instances",
        )

    primitive_instances: List[PrimitiveBase] = []
    for primitive in primitives:
        if inspect.isclass(primitive) and issubclass(primitive, PrimitiveBase):
            primitive_instances.append(primitive())
        elif isinstance(primitive, PrimitiveBase):
            primitive_instances.append(primitive)
        else:
            raise ValueError(
                "primitives must be a list of Primitive classes or Primitive instances",
            )

    return (input_features, primitive_instances)


def generate_features_from_primitives(
    input_features: Iterable[LiteFeature],
    primitives: Union[List[Type[PrimitiveBase]], List[PrimitiveBase]],
) -> List[LiteFeature]:
    """
    ** EXPERIMENTAL **
    Calculates all Features for a given input of features and a list of primitives.

    Args:
        origin_features (List[LiteFeature]):
            List of origin features
        primitives (List[Type[PrimitiveBase]])
            List of primitive classes

    Returns:
        List[LiteFeature]

    Examples:
        .. code-block:: python

            from featuretools.feature_discovery.feature_discovery import lite_dfs
            from featuretools.primitives import Absolute, IsNull
            import pandas as pd
            import woodwork as ww

            df = pd.DataFrame({
                "idx": [0,1,2,3],
                "f1": ["A", "B", "C", "D"],
                "f2": [1.2, 2.3, 3.4, 4.5]
            })

            df.ww.init()
            origin_features = schema_to_features(df.ww.schema)
            features = lite_dfs(origin_features, [Absolute, IsNull])

    """

    (input_features, primitives) = _check_inputs(input_features, primitives)

    features = [x.copy() for x in input_features]

    feature_collection = FeatureCollection(features=features)
    feature_collection.reindex()

    for primitive in primitives:
        features_ = _features_from_primitive(
            primitive=primitive,
            feature_collection=feature_collection,
        )
        features.extend(features_)

    return features



# flake8: noqa
from featuretools.feature_base.feature_base import (
    AggregationFeature,
    DirectFeature,
    Feature,
    FeatureBase,
    FeatureOutputSlice,
    GroupByTransformFeature,
    IdentityFeature,
    TransformFeature,
)
from featuretools.feature_base.feature_descriptions import describe_feature
from featuretools.feature_base.feature_visualizer import graph_feature
from featuretools.feature_base.features_deserializer import load_features
from featuretools.feature_base.features_serializer import save_features



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class IsIn(TransformPrimitive):
    """Determines whether a value is present in a provided list.

    Examples:
        >>> items = ['string', 10.3, False]
        >>> is_in = IsIn(list_of_outputs=items)
        >>> is_in(['string', 10.5, False]).tolist()
        [True, False, True]
    """

    name = "isin"
    input_types = [ColumnSchema()]
    return_type = ColumnSchema(logical_type=Boolean)
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]

    def __init__(self, list_of_outputs=None):
        self.list_of_outputs = list_of_outputs
        if not list_of_outputs:
            stringified_output_list = "[]"
        else:
            stringified_output_list = ", ".join([str(x) for x in list_of_outputs])
        self.description_template = "whether {{}} is in {}".format(
            stringified_output_list,
        )

    def get_function(self):
        def pd_is_in(array):
            return array.isin(self.list_of_outputs or [])

        return pd_is_in

    def generate_name(self, base_feature_names):
        return "%s.isin(%s)" % (base_feature_names[0], str(self.list_of_outputs))



import numpy as np
import pandas as pd
import pandas.api.types as pdtypes
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable, Datetime, Ordinal

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class GreaterThan(TransformPrimitive):
    """Determines if values in one list are greater than another list.

    Description:
        Given a list of values X and a list of values Y, determine
        whether each value in X is greater than each corresponding
        value in Y. Equal pairs will return `False`.

    Examples:
        >>> greater_than = GreaterThan()
        >>> greater_than([2, 1, 2], [1, 2, 2]).tolist()
        [True, False, False]
    """

    name = "greater_than"
    input_types = [
        [
            ColumnSchema(semantic_tags={"numeric"}),
            ColumnSchema(semantic_tags={"numeric"}),
        ],
        [ColumnSchema(logical_type=Datetime), ColumnSchema(logical_type=Datetime)],
        [ColumnSchema(logical_type=Ordinal), ColumnSchema(logical_type=Ordinal)],
    ]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    compatibility = [Library.PANDAS, Library.DASK]
    description_template = "whether {} is greater than {}"

    def get_function(self):
        def greater_than(val1, val2):
            val1_is_categorical = pdtypes.is_categorical_dtype(val1)
            val2_is_categorical = pdtypes.is_categorical_dtype(val2)
            if val1_is_categorical and val2_is_categorical:
                if not all(val1.cat.categories == val2.cat.categories):
                    return val1.where(pd.isnull, np.nan)
            elif val1_is_categorical or val2_is_categorical:
                # This can happen because CFM does not set proper dtypes for intermediate
                # features, so some agg features that should be Ordinal don't yet have correct type.
                return val1.where(pd.isnull, np.nan)
            return val1 > val2

        return greater_than

    def generate_name(self, base_feature_names):
        return "%s > %s" % (base_feature_names[0], base_feature_names[1])



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable

from featuretools.primitives.base import AggregationPrimitive


class MinCount(AggregationPrimitive):
    """Calculates the number of occurrences of the min value in a list

    Args:
        skipna (bool): Determines if to use NA/null values. Defaults to
            True to skip NA/null. If skipna is False, and there are NaN
            values in the array, the min will be NaN regardless of
            the other values, and NaN will be returned.

    Examples:
        >>> min_count = MinCount()
        >>> min_count([1, 2, 5, 1, 5, 3, 5])
        2

        You can optionally specify how to handle NaN values

        >>> min_count_skipna = MinCount(skipna=False)
        >>> min_count_skipna([1, 2, 5, 1, 5, 3, None])
        nan
    """

    name = "min_count"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})

    def __init__(self, skipna=True):
        self.skipna = skipna

    def get_function(self):
        def min_count(x):
            xmin = x.min(skipna=self.skipna)
            if np.isnan(xmin):
                return np.nan
            return x.eq(xmin).sum()

        return min_count



# put longer TLDs first to avoid catching a small part of a longer TLD and escape periods
COMMON_TLDS = [
    "management",
    "technology",
    "solutions",
    "delivery",
    "services",
    "software",
    "digital",
    "finance",
    "monster",
    "network",
    "support",
    "systems",
    "website",
    "agency",
    "design",
    "events",
    "global",
    "health",
    "online",
    "stream",
    "studio",
    "travel",
    "apple",
    "click",
    "cloud",
    "email",
    "games",
    "group",
    "media",
    "ninja",
    "press",
    "rocks",
    "space",
    "store",
    "today",
    "tools",
    "video",
    "works",
    "world",
    "aero",
    "arpa",
    "asia",
    "bank",
    "best",
    "blog",
    "buzz",
    "care",
    "casa",
    "chat",
    "club",
    "coop",
    "cyou",
    "desi",
    "farm",
    "goog",
    "guru",
    "host",
    "info",
    "jobs",
    "life",
    "link",
    "live",
    "mobi",
    "name",
    "news",
    "page",
    "plus",
    "shop",
    "site",
    "team",
    "tech",
    "work",
    "zone",
    "app",
    "aws",
    "bid",
    "biz",
    "box",
    "cam",
    "cat",
    "com",
    "dev",
    "edu",
    "eus",
    "fun",
    "gov",
    "icu",
    "int",
    "ltd",
    "mil",
    "net",
    "nyc",
    "one",
    "onl",
    "org",
    "ovh",
    "pro",
    "pub",
    "run",
    "sap",
    "top",
    "vip",
    "win",
    "xxx",
    "xyz",
    "ac",
    "ad",
    "ae",
    "ag",
    "ai",
    "al",
    "am",
    "ar",
    "at",
    "au",
    "az",
    "ba",
    "bd",
    "be",
    "bg",
    "br",
    "by",
    "bz",
    "ca",
    "cc",
    "cf",
    "ch",
    "cl",
    "cm",
    "cn",
    "co",
    "cr",
    "cu",
    "cx",
    "cy",
    "cz",
    "de",
    "dk",
    "do",
    "ec",
    "ee",
    "eg",
    "es",
    "eu",
    "fi",
    "fm",
    "fr",
    "ga",
    "ge",
    "gg",
    "gl",
    "gq",
    "gr",
    "gs",
    "gt",
    "hk",
    "hn",
    "hr",
    "hu",
    "id",
    "ie",
    "il",
    "im",
    "in",
    "io",
    "ir",
    "is",
    "it",
    "jo",
    "jp",
    "ke",
    "kh",
    "ki",
    "kr",
    "kw",
    "kz",
    "la",
    "lb",
    "li",
    "lk",
    "lt",
    "lu",
    "lv",
    "ly",
    "ma",
    "md",
    "me",
    "mk",
    "ml",
    "mm",
    "mn",
    "ms",
    "mu",
    "mx",
    "my",
    "nf",
    "ng",
    "nl",
    "no",
    "np",
    "nu",
    "nz",
    "om",
    "pa",
    "pe",
    "ph",
    "pk",
    "pl",
    "pr",
    "ps",
    "pt",
    "pw",
    "py",
    "qa",
    "re",
    "ro",
    "rs",
    "ru",
    "sa",
    "sc",
    "se",
    "sg",
    "sh",
    "si",
    "sk",
    "so",
    "st",
    "su",
    "sv",
    "sx",
    "th",
    "tj",
    "tk",
    "tn",
    "to",
    "tr",
    "tt",
    "tv",
    "tw",
    "ua",
    "ug",
    "uk",
    "us",
    "uy",
    "vc",
    "ve",
    "vn",
    "ws",
    "za",
]



import pandas as pd
import pytest

from featuretools.tests.testing_utils import make_ecommerce_entityset


def test_tokenize_entityset(pd_es, pd_int_es):
    pytest.importorskip("dask", reason="Dask not installed, skipping")
    from dask.base import tokenize

    dupe = make_ecommerce_entityset()

    # check identitcal entitysets hash to same token
    assert tokenize(pd_es) == tokenize(dupe)

    # not same if product relationship is missing
    productless = make_ecommerce_entityset()
    productless.relationships.pop()
    assert tokenize(pd_es) != tokenize(productless)

    # not same if integer entityset
    assert tokenize(pd_es) != tokenize(pd_int_es)

    # add row to cohorts
    cohorts_df = dupe["cohorts"]
    new_row = pd.DataFrame(
        data={
            "cohort": [2],
            "cohort_name": None,
            "cohort_end": [pd.Timestamp("2011-04-08 12:00:00")],
        },
        columns=["cohort", "cohort_name", "cohort_end"],
        index=[2],
    )
    more_cohorts = pd.concat([cohorts_df, new_row])
    dupe.replace_dataframe(dataframe_name="cohorts", df=more_cohorts)
    assert tokenize(pd_es) == tokenize(dupe)



import numpy as np
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class Absolute(TransformPrimitive):
    """Computes the absolute value of a number.

    Examples:
        >>> absolute = Absolute()
        >>> absolute([3.0, -5.0, -2.4]).tolist()
        [3.0, 5.0, 2.4]
    """

    name = "absolute"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the absolute value of {}"

    def get_function(self):
        return np.absolute



from string import punctuation
from typing import Iterable

import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable, NaturalLanguage

from featuretools.primitives.base import TransformPrimitive
from featuretools.primitives.standard.transform.natural_language.constants import (
    DELIMITERS,
    common_words_1000,
)


class NumberOfCommonWords(TransformPrimitive):
    """Determines the number of common words in a string.

    Description:
        Given string, determine the number of words that appear in a supplied word set.
        The word set defaults to nlp_primitives.constants.common_words_1000. The string
        is case insensitive. The word bank should consist of only lower case strings. If a string is
        missing, return `NaN`.

    Args:
        word_set (set, optional): The set of words to look for in the string. These
            words should all be lower case strings.
        delimiters_regex (str, optional): The regular expression used to determine
            what separates words. Defaults to whitespace characters.

    Examples:
        >>> x = ['Hey! This is some natural language', 'bacon, cheesburger, AND, fries', 'I! Am. A; duck?']
        >>> number_of_common_words = NumberOfCommonWords(word_set={'and', 'some', 'am', 'a', 'the', 'is', 'i'})
        >>> number_of_common_words(x).tolist()
        [2, 1, 3]

        >>> x = ['Hey! This is. some. natural language']
        >>> number_of_common_words = NumberOfCommonWords(word_set={'hey', 'is', 'some'}, delimiters_regex="[ .]")
        >>> number_of_common_words(x).tolist()
        [3]
    """

    name = "number_of_common_words"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})

    default_value = 0

    def __init__(
        self,
        word_set=common_words_1000,
        delimiters_regex=DELIMITERS,
    ):
        self.delimiters_regex = delimiters_regex
        self.word_set = word_set

    def get_function(self):
        def get_num_in_word_bank(words):
            if not isinstance(words, Iterable):
                return pd.NA
            num_common_words = 0
            for w in words:
                if (
                    w.lower().strip(punctuation) in self.word_set
                ):  # assumes word_set is all lowercase
                    num_common_words += 1
            return num_common_words

        def num_common_words(x):
            words = x.str.split(self.delimiters_regex)
            return words.apply(get_num_in_word_bank)

        return num_common_words



from featuretools.primitives.standard.transform.binary.add_numeric import AddNumeric
from featuretools.primitives.standard.transform.binary.add_numeric_scalar import (
    AddNumericScalar,
)
from featuretools.primitives.standard.transform.binary.and_primitive import And
from featuretools.primitives.standard.transform.binary.divide_by_feature import (
    DivideByFeature,
)
from featuretools.primitives.standard.transform.binary.divide_numeric import (
    DivideNumeric,
)
from featuretools.primitives.standard.transform.binary.divide_numeric_scalar import (
    DivideNumericScalar,
)
from featuretools.primitives.standard.transform.binary.equal import Equal
from featuretools.primitives.standard.transform.binary.equal_scalar import EqualScalar
from featuretools.primitives.standard.transform.binary.greater_than import GreaterThan
from featuretools.primitives.standard.transform.binary.greater_than_equal_to import (
    GreaterThanEqualTo,
)
from featuretools.primitives.standard.transform.binary.greater_than_equal_to_scalar import (
    GreaterThanEqualToScalar,
)
from featuretools.primitives.standard.transform.binary.greater_than_scalar import (
    GreaterThanScalar,
)
from featuretools.primitives.standard.transform.binary.less_than import LessThan
from featuretools.primitives.standard.transform.binary.less_than_equal_to import (
    LessThanEqualTo,
)
from featuretools.primitives.standard.transform.binary.less_than_equal_to_scalar import (
    LessThanEqualToScalar,
)
from featuretools.primitives.standard.transform.binary.less_than_scalar import (
    LessThanScalar,
)
from featuretools.primitives.standard.transform.binary.modulo_by_feature import (
    ModuloByFeature,
)
from featuretools.primitives.standard.transform.binary.modulo_numeric import (
    ModuloNumeric,
)
from featuretools.primitives.standard.transform.binary.modulo_numeric_scalar import (
    ModuloNumericScalar,
)
from featuretools.primitives.standard.transform.binary.multiply_boolean import (
    MultiplyBoolean,
)
from featuretools.primitives.standard.transform.binary.multiply_numeric import (
    MultiplyNumeric,
)
from featuretools.primitives.standard.transform.binary.multiply_numeric_boolean import (
    MultiplyNumericBoolean,
)
from featuretools.primitives.standard.transform.binary.multiply_numeric_scalar import (
    MultiplyNumericScalar,
)
from featuretools.primitives.standard.transform.binary.not_equal import NotEqual
from featuretools.primitives.standard.transform.binary.not_equal_scalar import (
    NotEqualScalar,
)
from featuretools.primitives.standard.transform.binary.or_primitive import Or
from featuretools.primitives.standard.transform.binary.scalar_subtract_numeric_feature import (
    ScalarSubtractNumericFeature,
)
from featuretools.primitives.standard.transform.binary.subtract_numeric import (
    SubtractNumeric,
)
from featuretools.primitives.standard.transform.binary.subtract_numeric_scalar import (
    SubtractNumericScalar,
)



import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Categorical, Datetime

from featuretools.primitives.base import TransformPrimitive
from featuretools.primitives.standard.transform.datetime.utils import HolidayUtil


class DateToHoliday(TransformPrimitive):
    """Transforms time of an instance into the holiday name, if there is one.

    Description:
        If there is no holiday, it returns `NaN`. Currently only works for the
        United States and Canada with dates between 1950 and 2100.

    Args:
        country (str): Country to use for determining Holidays.
            Default is 'US'. Should be one of the available countries here:
            https://github.com/dr-prodigy/python-holidays#available-countries

    Examples:
        >>> from datetime import datetime
        >>> date_to_holiday = DateToHoliday()
        >>> dates = pd.Series([datetime(2016, 1, 1),
        ...          datetime(2016, 2, 27),
        ...          datetime(2017, 5, 29, 10, 30, 5),
        ...          datetime(2018, 7, 4)])
        >>> date_to_holiday(dates).tolist()
        ["New Year's Day", nan, 'Memorial Day', 'Independence Day']

        We can also change the country.

        >>> date_to_holiday_canada = DateToHoliday(country='Canada')
        >>> dates = pd.Series([datetime(2016, 7, 1),
        ...          datetime(2016, 11, 15),
        ...          datetime(2017, 12, 26),
        ...          datetime(2018, 9, 3)])
        >>> date_to_holiday_canada(dates).tolist()
        ['Canada Day', nan, 'Boxing Day', 'Labour Day']
    """

    name = "date_to_holiday"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Categorical, semantic_tags={"category"})

    def __init__(self, country="US"):
        self.country = country
        self.holidayUtil = HolidayUtil(country)

    def get_function(self):
        def date_to_holiday(x):
            holiday_df = self.holidayUtil.to_df()
            df = pd.DataFrame({"date": x})
            df["date"] = df["date"].dt.date.astype("datetime64[ns]")

            df = df.merge(
                holiday_df,
                how="left",
                left_on="date",
                right_on="holiday_date",
            )
            return df.names.values

        return date_to_holiday



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Double

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class NaturalLogarithm(TransformPrimitive):
    """Computes the natural logarithm of a number.

    Examples:
        >>> log = NaturalLogarithm()
        >>> results = log([1.0, np.e]).tolist()
        >>> results = [round(x, 2) for x in results]
        >>> results
        [0.0, 1.0]
    """

    name = "natural_logarithm"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the natural logarithm of {}"

    def get_function(self):
        return np.log
